# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from inspect import signature
from typing import Callable, Dict, List, Literal, Optional, Tuple

from tqdm import tqdm
from wakepy import keep

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.geometries import Geometry
from ansys.simai.core.data.types import (
    Identifiable,
    NamedFile,
    get_id_from_identifiable,
    get_object_from_identifiable,
)
from ansys.simai.core.data.workspaces import Workspace
from ansys.simai.core.errors import InvalidArguments

logger = logging.getLogger(__name__)


class _OptimizationTrialRun(ComputableDataModel):
    """Provides the local representation of an optimization trial run object.

    The optimization trial run is an iteration of the optimization process.
    Each trial run tests a geometry and returns new parameters for the next geometry to try.
    """


# Undocumented for now, users don't really need to interact with it
class _OptimizationTrialRunDirectory(Directory[_OptimizationTrialRun]):
    _data_model = _OptimizationTrialRun

    def get(self, trial_run_id: str):
        """Get a specific trial run from the server."""
        return self._model_from(self._client._api.get_optimization_trial_run(trial_run_id))

    def _run_iteration(
        self, optimization: Identifiable["Optimization"], parameters: Dict
    ) -> _OptimizationTrialRun:
        optimization_id = get_id_from_identifiable(optimization)
        return self._model_from(
            self._client._api.run_optimization_trial(optimization_id, parameters)
        )


class Optimization(ComputableDataModel):
    """Provides the local representation of an optimization definition object."""

    def _run_iteration(self, parameters: Dict) -> "_OptimizationTrialRun":
        return self._client._optimization_trial_run_directory._run_iteration(self.id, parameters)


class OptimizationDirectory(Directory[Optimization]):
    """Provides a collection of methods related to optimizations.

    This class is accessed through ``client.optimizations``.

    Example:
        .. code-block:: python

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.optimizations.run(...)
    """

    _data_model = Optimization

    def get(self, optimization_id: str) -> Optimization:
        """Get a specific optimization object from the server.

        Args:
            optimization_id: ID of the optimization.

        Returns:
            :py:class:`Optimization`.
        """
        return self._model_from(self._client._api.get_optimization(optimization_id))

    def run_parametric(
        self,
        geometry_generation_fn: Callable[..., NamedFile],
        geometry_parameters: Dict[str, Tuple[float, float]],
        boundary_conditions: Dict[str, float],
        n_iters: int,
        minimize: Optional[List[str]] = None,
        maximize: Optional[List[str]] = None,
        outcome_constraints: Optional[List[str]] = None,
        show_progress: bool = False,
        workspace: Optional[Identifiable[Workspace]] = None,
    ) -> List[Dict]:
        """Run an optimization loop with client-side parametric geometry generation.

        Args:
            geometry_generation_fn: Function to call to generate a new geometry
                with the generated parameters. This parameter should return a
                :obj:`~ansys.simai.core.data.types.NamedFile` object.
            geometry_parameters: Name of the geometry parameters and their bounds or possible values (choices).
            boundary_conditions: Values of the boundary conditions to perform the optimization at.
                The values should map to existing boundary conditions in your project/workspace.
            n_iters: Number of iterations of the optimization loop.
            minimize: List of global coefficients to minimize.
                The global coefficients should map to existing coefficients in your project/workspace.
            maximize: List of global coefficients to maximize.
                The global coefficients should map to existing coefficients in your project/workspace.
            outcome_constraints: List of strings representing a linear inequality constraint
                on a global coefficient. The outcome constraint should be in the form ``gc >= x``,
                where:

                - ``gc`` is a valid global coefficient name.
                - ``x`` is a float bound.
                - The comparison operator is ``>=`` or ``<=``.

            show_progress: Whether to print progress on stdout.
            workspace: Workspace to run the optimization in. If a workspace is
                not specified, the default is the configured workspace.

        Returns:
            List of dictionaries representing the result of each iteration. when constraints
            are specified, the list can be shorter than the number of iterations.

        Warning:
            This is a long-running process and your computer must be powered on to generate the iterations.
            This method attempts to prevent your computer from sleeping, keeping your computer open
            during the process.

        Example:
          .. code-block:: python

            import ansys.simai.core


            # Function takes the parameters
            def my_geometry_generation_function(param_a, param_b):
                # Implementation
                return "/path/to/generated/geometry.stl"


            simai = ansys.simai.core.from_config(workspace="optimization-workspace")

            results = simai.optimizations.run(
                geometry_generation_fn=my_geometry_generation_function,
                geometry_parameters={
                    "param_a": {"bounds": (-12.5, 12.5)},
                    "param_b": {"choices": (0, 1)},
                },
                minimize=["TotalForceX"],
                boundary_conditions={"VelocityX": 10.5},
                outcome_constraints=["TotalForceY <= 10"],
                n_iters=100,
            )

            print(results)
        """
        workspace_id = get_id_from_identifiable(workspace, True, self._client._current_workspace)
        _validate_geometry_parameters(geometry_parameters)
        _validate_geometry_generation_fn_signature(geometry_generation_fn, geometry_parameters)
        objective = _build_objective(minimize, maximize)
        optimization_parameters = {
            "boundary_conditions": boundary_conditions,
            "n_iters": n_iters,
            "objective": objective,
            "type": "parametric",
            "outcome_constraints": outcome_constraints or [],
            "geometry_generation": {
                "geometry_parameters": geometry_parameters,
            },
        }
        with tqdm(total=n_iters, disable=not show_progress) as progress_bar:
            progress_bar.set_description("Creating optimization definition.")
            optimization = self._model_from(
                self._client._api.define_optimization(workspace_id, optimization_parameters)
            )
            optimization.wait()
            geometry_parameters = optimization.fields["initial_geometry_parameters"]
            logger.debug("Optimization defined. Starting optimization loop.")
            iterations_results: List[Dict] = []
            with keep.running(on_fail="warn"):
                while geometry_parameters:
                    logger.debug(f"Generating geometry with parameters {geometry_parameters}.")
                    progress_bar.set_description("Generating geometry.")
                    # TODO: Somehow keep session alive for long geometry generation
                    generated_geometry = geometry_generation_fn(**geometry_parameters)
                    logger.debug("Uploading geometry.")
                    progress_bar.set_description("Uploading geometry.")
                    # TODO: Name geometry ourselves ? Then we need to know the output format
                    geometry = self._client.geometries.upload(
                        generated_geometry,
                        metadata=geometry_parameters,
                        workspace_id=workspace_id,
                    )
                    logger.debug("Running trial.")
                    progress_bar.set_description("Running trial.")
                    trial_run = optimization._run_iteration(
                        {"geometry": geometry.id, "geometry_parameters": geometry_parameters}
                    )
                    trial_run.wait()
                    iteration_result = {
                        "parameters": geometry_parameters,
                        "objective": trial_run.fields["outcome_values"],
                    }
                    progress_bar.set_postfix(**iteration_result)
                    if trial_run.fields.get("is_feasible", True):
                        iterations_results.append(iteration_result)
                    else:
                        logger.debug("Trial run results did not match constraints. Skipping.")
                    geometry_parameters = trial_run.fields["next_geometry_parameters"]
                    logger.debug("Trial complete.")
                    progress_bar.update(1)
                logger.debug("Optimization complete.")
                progress_bar.set_description("Optimization complete.")
            return iterations_results

    def run_non_parametric(
        self,
        geometry: Identifiable[Geometry],
        bounding_boxes: List[List[float]],
        symmetries: List[Literal["x", "y", "z", "X", "Y", "Z"]],
        boundary_conditions: Dict[str, float],
        n_iters: int,
        minimize: Optional[List[str]] = None,
        maximize: Optional[List[str]] = None,
        show_progress: bool = False,
    ):
        """Run an optimization loop with server-side geometry generation using automorphism.

        Args:
            geometry: The base geometry on which to perform the automorphism. The optimization will
                run in the same workspace as the geometry.
            bounding_boxes: list of the bounds of the different boxes that will define the locations
                of the geometry to optimize.
                The format is [
                [box1_xmin, box1_xmax, box1_ymin, box1_ymax, box1_zmin, box1_zmax],
                [box2_xmin, box2_xmax, box2_ymin, box2_ymax, box2_zmin, box2_zmax],
                ...
                ]
            symmetries: list of symmetry axes, axes being x, y or z
            boundary_conditions: Values of the boundary conditions to perform the optimization at.
                The values should map to existing boundary conditions in your project/workspace.
            n_iters: Number of iterations of the optimization loop.
            minimize: List of global coefficients to minimize.
                The global coefficients should map to existing coefficients in your project/workspace.
            maximize: List of global coefficients to maximize.
                The global coefficients should map to existing coefficients in your project/workspace.
            show_progress: Whether to print progress on stdout.

        Warning:
            This feature is in beta. Results are not guaranteed.
        """
        geometry = get_object_from_identifiable(geometry, self._client._geometry_directory)
        objective = _build_objective(minimize, maximize)
        optimization_parameters = {
            "boundary_conditions": boundary_conditions,
            "n_iters": n_iters,
            "objective": objective,
            "type": "non_parametric",
            "geometry_generation": {
                "geometry": geometry.id,
                "box_bounds_list": bounding_boxes,
                "symmetries": symmetries,
            },
        }
        with tqdm(total=n_iters, disable=not show_progress) as progress_bar:
            progress_bar.set_description("Creating optimization definition.")
            optimization = self._model_from(
                self._client._api.define_optimization(
                    geometry._fields["workspace_id"], optimization_parameters
                )
            )
            optimization.wait()
            logger.debug("Optimization defined. Starting optimization loop.")
            iterations_results: List[Dict] = []
            with keep.running(on_fail="warn"):
                for _ in range(n_iters):
                    logger.debug("Running iteration")
                    progress_bar.set_description("Running iteration")
                    trial_run = optimization._run_iteration({})
                    trial_run.wait()
                    iteration_result = {
                        "objective": trial_run.fields["outcome_values"],
                    }
                    progress_bar.set_postfix(**iteration_result)
                    if trial_run.fields.get("is_feasible", True):
                        iterations_results.append(iteration_result)
                    else:
                        logger.debug("Trial run results did not match constraints. Skipping.")
                    logger.debug("Trial complete.")
                    progress_bar.update(1)
                logger.debug("Optimization complete.")
                progress_bar.set_description("Optimization complete.")
            return iterations_results


def _validate_geometry_parameters(params: Dict):
    if not isinstance(params, Dict):
        raise InvalidArguments("geometry_parameters: must be a dict.")
    if not params:
        raise InvalidArguments("geometry_parameters: must not be empty.")
    for key, value in params.items():
        bounds = value.get("bounds")
        choices = value.get("choices")
        if not bounds and not choices:
            raise InvalidArguments(f"geometry_parameters: no bounds or choices specified for {key}")
        if bounds and choices:
            raise InvalidArguments(
                f"geometry_parameters: only one of bounds or choices must be specified for {key}"
            )


def _validate_geometry_generation_fn_signature(geometry_generation_fn, geometry_parameters):
    geometry_generation_fn_args = signature(geometry_generation_fn).parameters
    if geometry_generation_fn_args.keys() != geometry_parameters.keys():
        raise InvalidArguments(
            f"geometry_generation_fn requires the following signature: {list(geometry_parameters.keys())}, but got: {list(geometry_generation_fn_args.keys())}"
        )


def _build_objective(minimize: List[str], maximize: List[str]) -> Dict:
    if not minimize and not maximize:
        raise InvalidArguments("No global coefficient to optimize.")
    objective = {}
    if minimize:
        for global_coefficient in minimize:
            objective[global_coefficient] = {"minimize": True}
    if maximize:
        for global_coefficient in maximize:
            objective[global_coefficient] = {"minimize": False}
    return objective



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, List, Literal, Optional

from ansys.simai.core.errors import InvalidArguments, ProcessingError
from ansys.simai.core.utils.misc import dict_get

if TYPE_CHECKING:
    from ansys.simai.core.data.projects import Project

SupportedBuildPresets = {
    "debug": "debug",
    "1_day": "short",
    "2_days": "default",
    "7_days": "long",
}


@dataclass
class DomainAxisDefinition:
    """Defines an axis in the Domain of Analysis.

    Args:
        position: Anchor point position.

                    | *relative_to_min*: `VolXmin = xmin - value`

                    | *relative_to_max*: `VolXmax = xmax + value`

                    | *relative_to_center*: `(xmin+xmax)/2 - value`

                    | *absolute*: `VolXmin = value`

        value:      Distance of the anchor from the position.
                        When ``position=absolute``, the distance can be either positive or negative.
                        In any other case, only positive values are accepted.
        length:     Length of the Domain of Analysis along the axis. Only positive numbers are accepted.

    Example:
        Define the Z-axis(i.e., height) in a Domain of Analysis

        .. code-block:: python

            from ansys.simai.core.data.models import DomainAxisDefinition, DomainOfAnalysis

            # Get the last configuration from a project
            bld_conf = my_project.last_model_configuration

            # Define a new axis for the Domain of Analysis
            new_doa_height = DomainAxisDefinition("relative_to_min", 180.5, 99.1)

            # Assign the new Domain of Analysis to the configuration
            bld_conf.domain_of_analysis.height = new_doa_height

    """

    def __post_init__(self):
        """Assess whether the value and length are correctly set."""

        self.__validate_length(self.length)
        self.__validate_value(self.value)

    def __set_value(self, val: float):
        self.__validate_value(val)
        self.__dict__["value"] = val

    def __get_value(self):
        return self.__dict__.get("value")

    def __validate_value(self, val: float):
        if val < 0 and self.position != "absolute":
            raise InvalidArguments(
                f"{self.__class__.__name__}: 'value' must be a positive number when the position is not 'absolute'.",
            )

    def __set_length(self, lgth: float):
        self.__validate_length(lgth)
        self.__dict__["length"] = lgth

    def __get_length(self):
        return self.__dict__.get("length")

    def __validate_length(self, lgth: float):
        if not lgth > 0:
            raise InvalidArguments(
                f"{self.__class__.__name__}: 'length' must be a positive number.",
            )

    position: Literal["relative_to_min", "relative_to_max", "relative_to_center", "absolute"]
    value: float = property(__get_value, __set_value)
    length: float = property(__get_length, __set_length)

    del __set_length, __set_value, __get_length, __get_value


@dataclass
class DomainOfAnalysis:
    """Defines the Domain of Analysis.

    Args:
        length: Domain of Analysis along the X axis
        width: Domain of Analysis along the Y axis
        height: Domain of Analysis along the Z axis

    Example:
        Get the Domain of Analysis from a configuration and replace it with a new one

        .. code-block:: python

            from ansys.simai.core.data.models import DomainOfAnalysis

            # Get the last configuration from a project
            bld_conf = my_project.last_model_configuration

            # Define a new Domain of Analysis
            new_doa = DomainOfAnalysis(
                length=("relative_to_max", 5, 8.1),
                width=("relative_to_max", 5, 8.1),
                height=("absolute", -4.5, 0.1),
            )

            # Assign the new Domain of Analysis to the configuration
            bld_conf.domain_of_analysis = new_doa

    """

    length: DomainAxisDefinition = None
    width: DomainAxisDefinition = None
    height: DomainAxisDefinition = None

    def __post_init__(self):
        if isinstance(self.length, tuple):
            self.length = DomainAxisDefinition(*self.length)
        if isinstance(self.width, tuple):
            self.width = DomainAxisDefinition(*self.width)
        if isinstance(self.height, tuple):
            self.height = DomainAxisDefinition(*self.height)


@dataclass
class GlobalCoefficientDefinition:
    """Global coefficient definition/parameter field.

    Args:
        formula: Global Coefficient formula.
        name: Global Coefficient name.
    """

    formula: str
    name: str


@dataclass
class ModelInput:
    """Model inputs.

    Args:
        surface: Input surface variables.
        boundary_conditions: Boundary conditions.
    """

    surface: list[str] = None
    boundary_conditions: list[str] = None


@dataclass
class ModelOutput:
    """The outputs of a model.

    Args:
        surface: the output surface variables.
        volume: the output volume variables.
    """

    surface: list[str] = None
    volume: list[str] = None


@dataclass
class PostProcessInput:
    """Designates the variables to use as post-processing input.

    Args:
        surface: the post-processing input surface variables.
    """

    surface: list[str] = None


@dataclass
class ModelConfiguration:
    """Configures the build of a model.

    Args:
        project: the project of the configuration.
        build_preset: indicates the build duration. Available options:

                    | *debug*: < 30 min, only 4 dat

                    | *1_day*: < 24 hours

                    | *2_days*: < 2 days, default value.

                    | *7_days*: < 1 week
        build_on_top: indicates if build_on_top learning is enabled. Default is False.
        input: the inputs of the model.
        output: the outputs of the model.
        global_coefficients: the Global Coefficients of the model.
        domain_of_analysis: The Domain of Analysis of the model configuration.
        pp_input: The post-processing input (e.g. a surface variable).

    Example:
        Define a new configuration and launch a build.

        .. code-block:: python

            import ansys.simai.core as asc
            from ansys.simai.core.data.model_configuration import (
                DomainAxisDefinition,
                DomainOfAnalysis,
                ModelConfiguration,
                ModelInput,
                ModelOutput,
                PostProcessInput,
            )

            simai = asc.from_config()

            # Get the project of interest
            aero_dyn_project = simai.projects.get(name="aero-dyn")

            # Define the inputs of the model
            model_input = ModelInput(surface=["Velocity"], boundary_conditions=["Vx"])

            # Define the outputs of the model
            model_output = ModelOutput(
                surface=["Pressure", "WallShearStress_0"], volume=["Velocity_0", "Pressure"]
            )

            # Define the surface post-processing input
            pp_input = PostProcessInput(surface=["Temperature_1"])

            # Define the model coefficients
            global_coefficients = [("max(Pressure)", "maxpress")]

            # Set the Domain of Analysis
            doa = DomainOfAnalysis(
                length=("relative_to_max", 5, 8.1),
                width=("relative_to_max", 5, 8.1),
                height=("absolute", -4.5, 0.1),
            )

            # Define the build configuration for the model
            new_conf = ModelConfiguration(
                project=aero_dyn_project,
                build_preset="debug",
                build_on_top=False,
                input=model_input,
                output=model_output,
                global_coefficients=global_coefficients,
                domain_of_analysis=doa,
                pp_input=pp_input,
            )

            # Launch a model build with the new configuration
            new_model = simai.models.build(new_conf)
    """

    project: "Optional[Project]" = None
    build_on_top: bool = False
    domain_of_analysis: DomainOfAnalysis = field(default_factory=lambda: DomainOfAnalysis())

    def __set_gc(self, gcs: list[GlobalCoefficientDefinition]):
        verified_gcs = []

        for gc in gcs:
            gc_unit = GlobalCoefficientDefinition(*gc) if isinstance(gc, tuple) else gc
            if self.project is None:
                raise ProcessingError(
                    f"{self.__class__.__name__}: a project must be defined for setting global coefficients."
                )

            self.project.verify_gc_formula(
                gc_unit.formula, self.input.boundary_conditions, self.output.surface
            )
            verified_gcs.append(gc_unit)
        self.__dict__["global_coefficients"] = verified_gcs

    def __get_gc(self):
        return self.__dict__.get("global_coefficients")

    global_coefficients: list[GlobalCoefficientDefinition] = property(__get_gc, __set_gc)

    def __validate_build_preset(self, val: str):
        if val not in SupportedBuildPresets:
            raise InvalidArguments(
                f"Invalid value for build_preset, build_preset should be one of {list(SupportedBuildPresets.keys())}"
            )

    def __set_build_preset(self, val: str):
        self.__validate_build_preset(val)
        self.__dict__["build_preset"] = val

    def __get_build_preset(self):
        return self.__dict__.get("build_preset")

    build_preset = property(__get_build_preset, __set_build_preset)

    def __validate_variables(self, vars_to_validate: list[str], var_type: str):
        sample_metadata = self.project.sample.fields.get("extracted_metadata")
        var_fields = dict_get(sample_metadata, var_type, "fields", default=[])
        var_fields_name = {fd.get("name") for fd in var_fields}
        unknown_variables = set(vars_to_validate) - var_fields_name
        if unknown_variables:
            raise ProcessingError(
                f"{self.__class__.__name__}: {var_type} variables {', '.join(unknown_variables)} do not exist in the reference sample."
            )

    def __validate_surface_variables(self, vars_to_validate: list[str]):
        sample_metadata = self.project.sample.fields.get("extracted_metadata")
        if not sample_metadata.get("surface"):
            raise ProcessingError(
                "No surface field is found in the reference sample. A surface field is required to use surface variables."
            )
        self.__validate_variables(vars_to_validate, "surface")

    def __validate_volume_variables(self, vars_to_validate: list[str]):
        sample_metadata = self.project.sample.fields.get("extracted_metadata")
        if not sample_metadata.get("volume"):
            raise ProcessingError(
                "No volume field is found in the reference sample. A volume field is required to use volume variables."
            )
        self.__validate_variables(vars_to_validate, "volume")

    def __set_input(self, model_input: ModelInput):
        if not model_input:
            raise InvalidArguments(
                "Invalid value for input; input should be an instance of ModelInput."
            )
        if model_input.surface:
            self.__validate_surface_variables(model_input.surface)
        self.__dict__["input"] = model_input

    def __get_input(self):
        return self.__dict__.get("input")

    input = property(__get_input, __set_input)

    def __set_output(self, model_output: ModelOutput):
        if not model_output:
            raise InvalidArguments(
                "Invalid value for output; output should be an instance of ModelOutput."
            )
        if model_output.surface:
            self.__validate_surface_variables(model_output.surface)

        if model_output.volume:
            self.__validate_volume_variables(model_output.volume)

        self.__dict__["output"] = model_output

    def __get_output(self):
        return self.__dict__.get("output")

    output = property(__get_output, __set_output)

    def __set_pp_input(self, pp_input: PostProcessInput):
        if not pp_input:
            pp_input = PostProcessInput()
        if pp_input.surface:
            self.__validate_surface_variables(pp_input.surface)
        self.__dict__["pp_input"] = pp_input

    def __get_pp_input(self):
        return self.__dict__.get("pp_input")

    pp_input = property(__get_pp_input, __set_pp_input)

    def __init__(
        self,
        project: "Project",
        boundary_conditions: Optional[dict[str, Any]] = None,
        build_preset: Optional[str] = "debug",
        build_on_top: bool = False,
        fields: Optional[dict[str, Any]] = None,
        global_coefficients: Optional[list[GlobalCoefficientDefinition]] = None,
        simulation_volume: Optional[dict[str, Any]] = None,
        input: Optional[ModelInput] = None,
        output: Optional[ModelOutput] = None,
        domain_of_analysis: Optional[DomainOfAnalysis] = None,
        pp_input: Optional[PostProcessInput] = None,
    ):
        """Sets the properties of a build configuration."""
        self.project = project
        self.input = ModelInput()
        if input is not None:
            self.input = input
        self.output = ModelOutput()
        if output is not None:
            self.output = output
        self.pp_input = PostProcessInput()
        if pp_input is not None:
            self.pp_input = pp_input
        if boundary_conditions is not None and self.input.boundary_conditions is None:
            self.input.boundary_conditions = list(boundary_conditions.keys())
        self.build_preset = build_preset
        self.build_on_top = build_on_top
        if fields is not None:
            if fields.get("surface_input"):
                self.input.surface = [fd.get("name") for fd in fields["surface_input"]]

            if fields.get("surface"):
                self.output.surface = [fd.get("name") for fd in fields["surface"]]

            if fields.get("volume"):
                self.output.volume = [fd.get("name") for fd in fields["volume"]]

            if fields.get("surface_pp_input"):
                self.pp_input.surface = [fd.get("name") for fd in fields["surface_pp_input"]]

        self.domain_of_analysis = domain_of_analysis
        if simulation_volume is not None:
            self.domain_of_analysis = DomainOfAnalysis(
                length=self._get_doa_axis(simulation_volume, "X"),
                width=self._get_doa_axis(simulation_volume, "Y"),
                height=self._get_doa_axis(simulation_volume, "Z"),
            )

        if global_coefficients is not None:
            gcs = [
                GlobalCoefficientDefinition(**gc) if isinstance(gc, dict) else gc
                for gc in global_coefficients
            ]
            self.global_coefficients = gcs

    def _get_doa_axis(self, sim_vol: dict, rel_pos: str) -> DomainAxisDefinition:
        """Composes a DomainAxisDefinition from raw json."""
        pos = sim_vol.get(rel_pos)
        return DomainAxisDefinition(position=pos["type"], value=pos["value"], length=pos["length"])

    def _set_doa_axis(self, fld: DomainAxisDefinition, param: str) -> dict[str, Any]:
        """Serializes a DomainAxisDefinition to the required format for the server."""
        if fld is None:
            raise InvalidArguments(
                f"Empty parameter '{param}' found when setting the Domain of Analysis. All parameters should be set."
            )
        return {"length": fld.length, "type": fld.position, "value": fld.value}

    def _to_payload(self):
        """Constructs the payload for a build request."""

        bcs = {}
        if self.input.boundary_conditions is not None:
            bcs = {bc_name: {} for bc_name in self.input.boundary_conditions}

        sample_metadata = self.project.sample.fields.get("extracted_metadata")
        surface_fields = dict_get(sample_metadata, "surface", "fields", default=[])
        volume_fields = dict_get(sample_metadata, "volume", "fields", default=[])

        surface_input_fld = []
        if self.input.surface is not None:
            surface_input_fld = [
                fd for fd in surface_fields if fd.get("name") in self.input.surface
            ]

        surface_fld = []
        if self.output.surface is not None:
            surface_fld = [fd for fd in surface_fields if fd.get("name") in self.output.surface]

        volume_fld = []
        if self.output.volume:
            volume_fld = [fd for fd in volume_fields if fd.get("name") in self.output.volume]

        gcs = []
        if self.global_coefficients is not None:
            gcs = [asdict(gc) for gc in self.global_coefficients]

        surface_pp_input_fld = []
        if self.pp_input.surface is not None:
            surface_pp_input_fld = [
                fd for fd in surface_fields if fd.get("name") in self.pp_input.surface
            ]

        flds = {
            "surface": surface_fld,
            "surface_input": surface_input_fld,
            "volume": volume_fld,
            "surface_pp_input": surface_pp_input_fld,
        }

        simulation_volume = {
            "X": self._set_doa_axis(self.domain_of_analysis.length, "length"),
            "Y": self._set_doa_axis(self.domain_of_analysis.width, "width"),
            "Z": self._set_doa_axis(self.domain_of_analysis.height, "height"),
        }

        return {
            "boundary_conditions": bcs,
            "build_preset": SupportedBuildPresets[self.build_preset],
            "continuous": self.build_on_top,
            "fields": flds,
            "global_coefficients": gcs,
            "simulation_volume": simulation_volume,
        }

    @classmethod
    def _from_payload(cls, **kwargs) -> "ModelConfiguration":
        # Retrieve SDK version of build preset from API version of build preset
        build_preset = next(
            (k for k, v in SupportedBuildPresets.items() if v == kwargs.get("build_preset")), None
        )
        kwargs["build_preset"] = build_preset
        if build_on_top := kwargs.pop("continuous", None):
            kwargs["build_on_top"] = build_on_top
        return ModelConfiguration(**kwargs)

    def compute_global_coefficient(self) -> List[float]:
        """Computes the results of the formula for all global coefficients with respect to the project's sample."""

        if self.project is None:
            raise ProcessingError(
                f"{self.__class__.__name__}: a project must be a defined for computing the global coefficient formula."
            )

        return [
            self.project.compute_gc_formula(
                gc.formula, self.input.boundary_conditions, self.output.surface
            )
            for gc in self.global_coefficients
        ]

    del __set_build_preset, __get_build_preset



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
import itertools
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

from ansys.simai.core.data.types import Range, _HollowRange
from ansys.simai.core.utils.grouping import _ToleranceGrouper
from ansys.simai.core.utils.numerical import (
    DEFAULT_COMPARISON_EPSILON,
    is_equal_with_tolerance,
    is_number,
    is_smaller_with_tolerance,
    validate_tolerance_parameter,
)
from ansys.simai.core.utils.validation import _enforce_as_list_passing_predicate

if TYPE_CHECKING:
    from ansys.simai.core.data.geometries import Geometry


def _sweep(
    candidate_geometry: "Geometry",
    geometries: List["Geometry"],
    swept_metadata: Optional[Union[str, List[str]]] = None,
    fixed_metadata: Optional[List[str]] = None,
    order: Optional[int] = None,
    include_center: Optional[bool] = None,
    include_diagonals: Optional[bool] = None,
    tolerance: Optional[float] = None,
) -> List["Geometry"]:
    if order is None:
        order = 1
    if not isinstance(order, int):
        raise TypeError(f"The 'order' argument must be a positive integer (passed {order}).")
    if order < 1:
        raise ValueError(f"The 'order' argument must be a positive integer (passed {order}).")
    if tolerance is None:
        tolerance = DEFAULT_COMPARISON_EPSILON
    validate_tolerance_parameter(tolerance)

    if include_center is None:
        include_center = False
    if include_diagonals is None:
        include_diagonals = False

    if swept_metadata:
        swept_metadata = _enforce_as_list_passing_predicate(
            swept_metadata,
            lambda s: isinstance(s, str),
            """The 'swept_metadata' argument must be a geometry metadata name
            or a list of metadata names.
            """,
        )

        # Make sure all passed variables exist in geometry metadata
        # and that the data is numerical
        for variable in swept_metadata:
            if variable not in candidate_geometry.metadata:
                raise ValueError(f"Metadata {variable} not found in candidate geometry")
            if not is_number(candidate_geometry.metadata[variable]):
                raise ValueError(f"Metadata {variable} is not numerical")
    else:
        # No variable passed: use all numerical axes
        swept_metadata = [
            name
            for name in candidate_geometry.metadata
            if is_number(candidate_geometry.metadata[name])
        ]
        if len(swept_metadata) == 0:
            raise ValueError("No numerical metadata column found in candidate geometry")

    if fixed_metadata is None:
        fixed_metadata = []
    # consider swept variables as fixed
    eventually_fixed_metadata = list(set(fixed_metadata + swept_metadata))

    # Collect geometries in a set,
    # as when using include_center, a same geometry can be collected
    # via multiple variables
    neighboring_geometries: Set["Geometry"] = set()

    # if include_diagonal, collect boundaries
    if include_diagonals:
        boundaries: Dict[str, Tuple[float, float]] = {}

    # Go through each variable to find neighbors for those
    for checked_variable in swept_metadata:
        candidate_value: float = candidate_geometry.metadata[checked_variable]  # type: ignore
        fixed_variables = [v for v in eventually_fixed_metadata if v != checked_variable]
        # 1. Find geometries whose metadata contain checked_variable
        # and match fixed values
        matching_geometries = [
            geometry
            for geometry in geometries
            if (
                checked_variable in geometry.metadata
                and _are_geometries_metadata_equal(
                    left_geometry=candidate_geometry,
                    right_geometry=geometry,
                    compared_variables=fixed_variables,
                )
            )
        ]
        # 2. Sort them by checked_variable, so they can be passed to groupby
        sorted_geometries: List[Geometry] = sorted(
            matching_geometries, key=lambda g: g.metadata[checked_variable]
        )
        # 2. Group into buckets of equal value.
        # Use itertools.groupby to group.
        # As we want equality with a tolerance,
        # also that candidate geometry is at the center of its bucket,
        # we need this custom comparator
        groupby_bucketizer = _ToleranceGrouper(
            # FIXME: fix the B023 issue
            key_func=lambda g: g.metadata[checked_variable],  # noqa: B023
            tolerance=tolerance,
            forced_central_value=candidate_value,
        )
        buckets = itertools.groupby(sorted_geometries, groupby_bucketizer)

        # 3. Now keep only as many smaller and bigger buckets as `order`.
        # Keep the central values if include_center is True.

        # Smaller neighbors are selected by a deque of max size=`order`:
        # we push in it all the smaller values; the deque
        # keeps only as many as we want (nb=`order`)
        smaller_groups = collections.deque([], order)
        # Bigger_geometries are collected in a list by counting directly
        equal_and_bigger_groups = []
        found_bigger_buckets = 0
        for bucket, geometry_group in buckets:
            if is_smaller_with_tolerance(bucket.center, candidate_value, tolerance=tolerance):
                # smaller: add to the deque
                smaller_groups.append((bucket, list(geometry_group)))
            elif is_equal_with_tolerance(bucket.center, candidate_value, tolerance=tolerance):
                # equal: only add if include_center
                if include_center:
                    equal_and_bigger_groups.append((bucket, list(geometry_group)))
            else:
                # bigger: add to equal_and_bigger_groups
                equal_and_bigger_groups.append((bucket, list(geometry_group)))
                found_bigger_buckets += 1
                if found_bigger_buckets >= order:
                    # we found as many bigger as we were looking for: stop
                    break
        # Store everything in neighboring_geometries
        if include_diagonals:
            min_boundary = None
            max_boundary = None
        for _bucket, geometry_group in itertools.chain(smaller_groups, equal_and_bigger_groups):
            neighboring_geometries.update(geometry_group)
            if include_diagonals:
                if min_boundary is None:
                    min_boundary = _bucket.center
                max_boundary = _bucket.center
        if include_diagonals:
            boundaries[checked_variable] = (min_boundary, max_boundary)
    if include_diagonals:
        # build the list of ranges matching all boundaries.
        # selected geometries must match all those ranges.
        all_ranges: Dict[str, Range] = {}
        for variable, (min_boundary, max_boundary) in boundaries.items():
            # If not include_center, we want the geometries
            # *not* to match the candidate's value.
            # Thus use the HollowRange to exclude it
            excluded_value: Optional[float] = None
            if not include_center:
                excluded_value = candidate_geometry.metadata[variable]
            all_ranges[variable] = _HollowRange(
                min=min_boundary,
                max=max_boundary,
                tolerance=tolerance,
                excluded_value=excluded_value,
            )

        # We loop again through all geometries
        # and select those matching both all limits on swept geometries
        # and fixed value
        for geometry in geometries:
            if geometry in neighboring_geometries:
                continue
            if not _geometry_matches_range_constraints(geometry, all_ranges):
                continue
            if not _are_geometries_metadata_equal(
                left_geometry=candidate_geometry,
                right_geometry=geometry,
                compared_variables=fixed_metadata,
            ):
                continue
            neighboring_geometries.add(geometry)

    return list(neighboring_geometries)


def _are_geometries_metadata_equal(
    left_geometry: "Geometry",
    right_geometry: "Geometry",
    compared_variables: List[str],
    tolerance: Optional[float] = None,
) -> bool:
    """Test if all compared metadata are equal between two geometries.

    If a metadata is absent, consider the comparison as false.

    Returns:
        ``True`` if the metadata are equal for considered columns.

    Raises:
        ValueError: If one considered column contains numerical
            data and the other column contains non-numerical data.
    """
    for variable in compared_variables:
        if variable not in left_geometry.metadata or variable not in right_geometry.metadata:
            return False
        left_val: float = left_geometry.metadata.get(variable)  # type: ignore
        right_val: float = right_geometry.metadata.get(variable)  # type: ignore
        if is_number(left_val) != is_number(right_val):
            raise ValueError("Mixed numerical and non-numerical values in metadata {variable}.")
        if is_number(left_val):
            if not is_equal_with_tolerance(left_val, right_val, tolerance=tolerance):
                return False
        else:
            if left_val != right_val:
                return False
    return True


def _geometry_matches_range_constraints(
    geometry: "Geometry", ranges: Optional[Dict[str, Range]] = None
) -> bool:
    if ranges is None:
        return True
    for metadata_name, range in ranges.items():
        value = geometry.metadata.get(metadata_name, None)
        if value is None:
            # this metadata is not filled for this geometry: no match
            return False
        if not is_number(value):
            # metadata is filled but is not a number: raise
            raise TypeError(
                """
                Range constraint can only be used
                on numerical metadata."""
            )
        if not range.match_value(value):
            return False
    return True



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import pathlib
import platform
import time
from typing import IO, TYPE_CHECKING, Any

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ansys.simai.core.data.types import Path


def _expand_user_path(file_path: "Path") -> pathlib.Path:
    """Convert string inputs to ``Path`` and expand the user.

    This method supports paths starting with ``~`` on Linux.
    """
    return pathlib.Path(str(file_path)).expanduser()


def file_path_to_obj_file(file_path: "Path", mode: str) -> IO[Any]:
    """Take a file path and return a file-object opened in the given mode."""
    file_path = _expand_user_path(file_path)
    file_path.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Opening file {file_path}")
    return open(file_path, mode=mode)  # noqa: SIM115


def get_cache_dir() -> pathlib.Path:
    system = platform.system()
    if system == "Windows":
        cache_dir = pathlib.Path(os.getenv("APPDATA", "~")) / "Ansys" / "cache" / "pysimai"
    elif system == "Linux":
        cache_dir = pathlib.Path(os.getenv("XDG_CACHE_HOME", "~/.cache")) / "ansys" / "pysimai"
    elif system == "Darwin":
        cache_dir = pathlib.Path("~/Library/Caches/Ansys/pysimai")
    else:
        raise RuntimeError(f"Unknown OS: {system}")
    cache_dir = cache_dir.expanduser()
    cache_dir.mkdir(exist_ok=True, parents=True)
    a_week_ago = time.time() - 7 * 86400
    for cache_entry in pathlib.Path(cache_dir).glob("*"):
        if cache_entry.is_file():
            itemTime = cache_entry.stat().st_mtime
            if itemTime < a_week_ago:
                cache_entry.unlink()
    return cache_dir



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from ansys.simai.core.data.lists import ExportablePPList, PPList
from ansys.simai.core.data.post_processings import (
    GlobalCoefficients,
    Slice,
    SurfaceEvolution,
    SurfaceVTP,
    SurfaceVTPTDLocation,
    VolumeVTU,
)

if TYPE_CHECKING:
    from ansys.simai.core.data.selections import Selection


class SelectionPostProcessingsMethods:
    """Acts as a namespace inside :py:class:`~ansys.simai.core.data.selections.Selection` objects,
    allowing you to access or run postprocessings on whole selections.
    """

    def __init__(self, selection: "Selection"):
        self._selection = selection

    def global_coefficients(self) -> ExportablePPList[GlobalCoefficients]:
        """Compute or get the global coefficients of the selected predictions.

        This is a non-blocking method. It returns an
        :py:class:`~ansys.simai.core.data.lists.ExportablePPList` instance
        of :py:class:`~ansys.simai.core.data.post_processings.GlobalCoefficients`
        objects without waiting. Those ``PostProcessing`` objects may not have
        data right away if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be waited upon with the ``wait()`` method.

        The computation is launched only on the first call of this method.
        Subsequent calls do not relaunch it.

        Returns:
            :py:class:`~ansys.simai.core.data.lists.ExportablePPList` instance
            of :py:class:`~ansys.simai.core.data.post_processings.GlobalCoefficients`
            objects.
        """
        return ExportablePPList(
            selection=self._selection, post=lambda pred: pred.post.global_coefficients()
        )

    def surface_evolution(self, axis: str, delta: float) -> ExportablePPList[SurfaceEvolution]:
        """Compute or get the SurfaceEvolution of the predictions for specific parameters.

        This is a non-blocking method. It returns an
        :py:class:`~ansys.simai.core.data.lists.ExportablePPList` instance
        of :py:class:`~ansys.simai.core.data.post_processings.SurfaceEvolution`
        objects without waiting. Those ``PostProcessing`` objects may not have
        data right away if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be waited upon with the ``wait()`` method.

        The computation is launched only on the first call of this method
        with a specific set of parameters.
        Subsequent calls with the same parameters do not relaunch it.

        Args:
            axis: Axis to compute the the SurfaceEvolution on.
            delta: Increment of the abscissa in meters.

        Returns:
            :py:class:`~ansys.simai.core.data.lists.ExportablePPList` instance of
            :py:class:`~ansys.simai.core.data.post_processings.SurfaceEvolution` objects.
        """
        return ExportablePPList(
            selection=self._selection,
            post=lambda pred: pred.post.surface_evolution(axis, delta),
        )

    def slice(self, axis: str, coordinate: float) -> PPList[Slice]:
        """Compute or get a slice from each prediction in the selection.

        This is a non-blocking method. It returns a
        :py:class:`~ansys.simai.core.data.lists.PPList` instance of
        :py:class:`~ansys.simai.core.data.post_processings.Slice`
        objects without waiting. Those ``PostProcessing`` objects may not have
        data right away if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be waited upon with the ``wait()`` method.

        The computation is launched only on the first call of this method
        with a specific set of parameters.
        Subsequent calls with the same parameters do not relaunch it.

        The slices are in the NPZ format.

        Args:
            axis: Axis to slice.
            coordinate: Coordinate along the given axis to slice at.

        Returns:
            :py:class:`~ansys.simai.core.data.lists.PPList` list of
            :py:class:`~ansys.simai.core.data.post_processings.Slice` objects.
        """
        return PPList(
            selection=self._selection,
            post=lambda pred: pred.post.slice(axis=axis, coordinate=coordinate),
        )

    def volume_vtu(self) -> PPList[VolumeVTU]:
        """Compute or get the result of each prediction's volume in the VTU format.

        This is a non-blocking method. It returns a
        :py:class:`~ansys.simai.core.data.lists.PPList` instance of
        :py:class:`~ansys.simai.core.data.post_processings.VolumeVTU`
        objects without waiting. Those ``PostProcessing`` objects may not have
        data right away if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be waited upon with the ``wait()`` method.

        The computation is launched only on the first call of this method.
        Subsequent calls do not relaunch it.

        Returns:
            :py:class:`~ansys.simai.core.data.lists.PPList` instance of
            :py:class:`~ansys.simai.core.data.post_processings.VolumeVTU` objects.
        """
        return PPList(selection=self._selection, post=lambda pred: pred.post.volume_vtu())

    def surface_vtp(self) -> PPList[SurfaceVTP]:
        """Compute or get the result of each prediction's surface in the VTP format.

        This is a non-blocking method. It returns a
        :py:class:`~ansys.simai.core.data.lists.PPList` instance of
        :py:class:`~ansys.simai.core.data.post_processings.SurfaceVTP`
        objects without waiting. Those ``PostProcessing`` objects may not have
        data right away if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be waited upon with the ``wait()`` method.


        The computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Returns:
            :py:class:`~ansys.simai.core.data.lists.PPList` instance of
            :py:class:`~ansys.simai.core.data.post_processings.SurfaceVTP` objects.
        """
        return PPList(selection=self._selection, post=lambda pred: pred.post.surface_vtp())

    def surface_vtp_td_location(self) -> PPList[SurfaceVTPTDLocation]:
        """Compute or get the result of each prediction's surface in the VTP format.

        This method keeps the original data association as they are in the sample.

        It is a non-blocking method. It returns a
        :py:class:`~ansys.simai.core.data.lists.PPList` instance of
        :py:class:`~ansys.simai.core.data.post_processings.SurfaceVTPTDLocation`,
        objects without waiting. Those ``PostProcessing`` objects may not have
        data right away if the computation is still in progress. Data is filled
        asynchronously once the computation is finished.
        The state of computation can be waited upon with the ``wait()`` method.


        The computation is launched only on first call of this method.
        Subsequent calls do not relaunch it.

        Returns:
            :py:class:`~ansys.simai.core.data.lists.PPList` instance of
            :py:class:`~ansys.simai.core.data.post_processings.SurfaceVTPTDLocation`
        """
        return PPList(
            selection=self._selection, post=lambda pred: pred.post.surface_vtp_td_location()
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Callable, List, TypeVar, Union

T = TypeVar("T")


def _list_elements_pass_predicate(items_list: List[T], predicate: Callable[[Any], bool]) -> bool:
    return isinstance(items_list, List) and all(predicate(item) for item in items_list)


def _enforce_as_list_passing_predicate(
    parameter: Union[T, List[T]], predicate: Callable[[Any], bool], error_message: str
) -> List[T]:
    """Make sure that the passed parameter is either a single element passing a predicate
    or a list of elements that are all passing predicates.

    In both case, a list is returned. In other cases, raise a TypeError with the
    error message.

    This method is useful for validating a type of parameter. For example, accepting either
    a geometry or a list of geometries.
    """
    if predicate(parameter):
        return [parameter]
    if _list_elements_pass_predicate(parameter, predicate):
        return parameter
    raise TypeError(error_message)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Dict, Iterator
from urllib.parse import quote

from ansys.simai.core.api.mixin import ApiClientMixin


class ProjectClientMixin(ApiClientMixin):
    def projects(self):
        return self._get("projects")

    def get_project(self, id: str):
        return self._get(f"projects/{id}")

    def get_project_by_name(self, name: str):
        return self._get(f"projects/name/{quote(name)}")

    def create_project(self, **kwargs):
        return self._post("projects", json=kwargs)

    def update_project(self, project_id: str, name: str):
        """Update a project name.

        Args:
            project_id: ID of the project.
            name: New name to give to the project.
        """
        request_json = {}
        request_json["name"] = name
        self._patch(f"projects/{project_id}", json=request_json, return_json=False)

    def iter_training_data_in_project(self, project_id: str) -> Iterator[Dict[str, Any]]:
        next_page = f"projects/{project_id}/data"
        while next_page:
            page_request = self._get(next_page, return_json=False)
            next_page = page_request.links.get("next", {}).get("url")
            yield from page_request.json()

    def set_project_sample(self, project_id: str, training_data_id: str):
        self._put(
            f"projects/{project_id}/sample",
            json={"training_data": training_data_id},
            return_json=False,
        )

    def delete_project(self, project_id: str):
        self._delete(f"projects/{project_id}", return_json=False)

    def launch_build(
        self,
        project_id: str,
        config: Dict[str, Any],
        dismiss_data_with_fields_discrepancies: bool = False,
        dismiss_data_with_volume_overflow: bool = False,
    ):
        """Launches a build for a project and according to a given configuration.

        Args:
            project_id: the ID of the project
            config: the build configuration
            dismiss_data_with_fields_discrepancies: set to True for omitting data with missing properties
            dismiss_data_with_volume_overflow: set to True for omitting data outside the Domain of Analysis

        """
        params = {
            "dismiss_data_with_fields_discrepancies": dismiss_data_with_fields_discrepancies,
            "dismiss_data_with_volume_overflow": dismiss_data_with_volume_overflow,
        }
        return self._post(f"projects/{project_id}/model", json=config, params=params)

    def is_project_trainable(self, project_id: str):
        return self._get(f"projects/{project_id}/trainable")

    def check_formula(self, project_id: str, calculette_payload: Dict[str, Any]):
        """Verify the validity of a Global Coefficient formula.

        Args:
            project_id:         the ID of the project
            calculette_payload: the payload for calculette that includes the
                                formula that describes a Global Coefficient
        """

        return self._post(
            f"projects/{project_id}/check-formula",
            json=calculette_payload,
            return_json=False,
        )

    def compute_formula(self, project_id: str, calculette_payload: Dict[str, Any]):
        """Computes the result of a Global Coefficient formula according to the project's sample.

        Args:
            project_id:         the ID of the project
            calculette_payload: the payload for calculette that includes the
                                formula that describes a Global Coefficient
        """

        return self._post(
            f"projects/{project_id}/compute-formula",
            json=calculette_payload,
            return_json=False,
        )

    def cancel_build(self, project_id: str):
        """Cancels an existing model build if it exists.

        Args:
            project_id: ID of the project.
        """

        return self._post(f"projects/{project_id}/cancel-training", return_json=True)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Dict, List

from ansys.simai.core.api.mixin import ApiClientMixin
from ansys.simai.core.data.types import APIResponse


class PointCloudClientMixin(ApiClientMixin):
    def create_point_cloud(self, geometry_id: str, name: str, extension: str) -> APIResponse:
        """Create a point cloud without pushing the data.

        Args:
            geometry_id: ID of the geometry to assign the point cloud to.
            name: Name to give to the geometry.
            extension: Extension to give to the file.
        """
        post_data = {"name": name, "file_extension": extension}
        return self._post(f"geometries/{geometry_id}/point-cloud", json=post_data)

    def complete_point_cloud_upload(
        self, point_cloud_id: str, upload_id: str, parts: List[Dict[str, Any]]
    ):
        """Complete the upload of a point cloud.

        Args:
           point_cloud_id: ID of the point cloud to complete
           upload_id: ID used to upload the point cloud
           parts: List of the uploaded file parts
        """
        self._post(
            f"point-clouds/{point_cloud_id}/complete",
            json={"upload_id": upload_id, "parts": parts},
            return_json=False,
        )

    def delete_point_cloud(self, point_cloud_id: str):
        """Delete the specific point cloud file.

        Args:
            point_cloud_id: ID of the input cloud to delete
        """
        self._delete(f"point-clouds/{point_cloud_id}", return_json=False)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging

from ansys.simai.core.api.mixin import ApiClientMixin

logger = logging.getLogger(__name__)


class PredictionClientMixin(ApiClientMixin):
    """Provides the client for the Prediction ("/predictions") part of the API."""

    def predictions(self, workspace_id: str):
        """Get a list of all predictions."""
        return self._get("predictions/", params={"workspace": workspace_id})

    def get_prediction(self, prediction_id: str):
        """Get information on a single prediction.

        Args:
            prediction_id: ID of the prediction.
        """
        return self._get(f"predictions/{prediction_id}")

    def delete_prediction(self, prediction_id: str):
        """Delete a single prediction.

        Args:
            prediction_id: ID of the prediction.
        """
        return self._delete(
            f"predictions/{prediction_id}",
            params={"confirm": True},
            return_json=False,
        )

    def run_prediction(self, geometry_id: str, **kwargs):  # noqa: D417
        """Run a prediction on a given geometry.

        Args:
            geometry_id: ID of the target geometry.

        Keyword Arguments:
            boundary_conditions dict: Constraints of the problem in dictionary form.
            tolerance float: Delta under which two boundary condition components
                are considered equal. The default is ``10**-6``.
        """
        return self._post(f"geometries/{geometry_id}/predictions", json=kwargs)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, Optional

from ansys.simai.core.data.types import File

if TYPE_CHECKING:
    import ansys.simai.core.client


class DownloadableResult:
    """Provides the object representing a result data for a postprocessing in binary format."""

    def __init__(
        self,
        download_url: Optional[str],
        client: "ansys.simai.core.client.SimAIClient",
        request_method: str = "GET",
        request_json_body: Optional[Dict[str, Any]] = None,
    ):
        self.url = download_url
        self._client = client
        self._request_method = request_method
        self._request_json_body = request_json_body

    def download(self, file: File) -> None:
        """Download the postprocessing data to the specified file or path.

        Args:
            file: Binary file-object or path of the file to download the data into.
        """
        self._download_file(self.url, file)

    def in_memory(self) -> BytesIO:
        """Load the postprocessing data in memory.

        Returns:
            :class:`io.BytesIO` object containing the postprocessing data.
        """
        return self._download_file(self.url)

    def _download_file(self, url: str, file: Optional[File] = None) -> Optional[BytesIO]:
        return self._client._api.download_file(
            url,
            file,
            request_method=self._request_method,
            request_json_body=self._request_json_body,
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Dict, Optional
from urllib.parse import quote

from ansys.simai.core.api.mixin import ApiClientMixin
from ansys.simai.core.data.types import File


class WorkspaceClientMixin(ApiClientMixin):
    """Provides the client for the Workspace ("/workspaces") part of the API."""

    def workspaces(self):
        """List all workspaces."""
        return self._get("workspaces/")

    def get_workspace(self, workspace_id: str) -> Dict[str, Any]:
        """Get information on a single workspace.

        Args:
            workspace_id: ID of the workspace.

        Returns:
            JSON representation of the workspace.

        Raises:
            ansys.simai.core.errors.NotFoundError: If no workspace with that ID exists.
            ansys.simai.core.errors.ApiClientError: On other HTTP errors.
        """
        return self._get(f"workspaces/{workspace_id}")

    def get_workspace_by_name(self, name: str):
        """Get information on a single workspace by name instead of ID.

        Args:
            name: Name of the workspace.
        """
        return self._get(f"workspaces/name/{quote(name)}")

    def get_workspace_model_manifest(self, workspace_id):
        """Get the public part of the manifest for the given workspace.

        Args:
            workspace_id: ID of the workspace.

        Raises:
            ansys.simai.core.errors.NotFoundError: If no workspace with that ID exists.
            ansys.simai.core.errors.ApiClientError: On other HTTP errors.
        """
        return self._get(f"workspaces/{workspace_id}/model/manifest/public")

    def create_workspace(self, name: str, model_id: str, **kwargs):
        """Create a workspace.

        Args:
            name: Name to give to the new workspace.
            model_id: ID of the model that the workspace is to use.
            **kwargs: Additional arguments for the workspace creation.

        Returns:
            JSON representation of the new workspace.
        """
        return self._post("workspaces/", json={"name": name, "model": model_id, **kwargs})

    def delete_workspace(self, workspace_id: str):
        """Delete a workspace.

        Args:
            workspace_id: ID of the workspace.

        Raises:
            ansys.simai.core.errors.NotFoundError: If no workspace with that ID exists.
            ansys.simai.core.errors.ApiClientError: On other HTTP errors.
        """
        return self._delete(f"workspaces/{workspace_id}")

    def download_workspace_model_evaluation_report(self, workspace_id: str, file: Optional[File]):
        return self.download_file(f"workspaces/{workspace_id}/model-evaluation-report", file)

    def download_workspace_mer_data(self, workspace_id: str, file: Optional[File]):
        return self.download_file(f"workspaces/{workspace_id}/mer-data", file)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import hashlib
import logging
from os import PathLike
from typing import Literal, Optional, Union
from urllib.parse import urlparse, urlunparse

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    Field,
    HttpUrl,
    ValidationInfo,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError

from ansys.simai.core.utils.misc import prompt_for_input

logger = logging.getLogger(__name__)


def prompt_if_interactive(interactive, **kwargs):
    """Raise an error or prompt for input according to _interactive_mode."""
    if not interactive:
        raise PydanticCustomError(
            "conf_param_missing", f"""Missing parameter "{kwargs["name"]}" from configuration"""
        )
    return prompt_for_input(**kwargs)


class Credentials(BaseModel, extra="forbid"):
    username: str  # the model validator will call prompt_for_interactive when interactive is on
    "Username: Required if :code:`Credentials` is defined, automatically prompted."
    password: str  # like above
    "Password: Required if :code:`Credentials` is defined, automatically prompted."
    totp: Optional[str] = None  # like the above
    "One-time password: required if :code:`totp_enabled=True`, automatically prompted."

    @model_validator(mode="before")
    @classmethod
    def prompt(cls, values, info):
        if "username" not in values:
            values["username"] = prompt_if_interactive(
                interactive=info.data["interactive"], name="username"
            )
        if "password" not in values:
            values["password"] = prompt_if_interactive(
                interactive=info.data["interactive"],
                name="password",
                hide_input=True,
            )
        if values.pop("totp_enabled", False) and "totp" not in values:
            values["totp"] = prompt_if_interactive(
                interactive=info.data["interactive"], name="totp"
            )
        return values


class ClientConfig(BaseModel, extra="allow"):
    interactive: Optional[bool] = True
    "If True, it enables interaction with the terminal."
    url: HttpUrl = HttpUrl("https://api.simai.ansys.com/v2/")
    "URL to the SimAI API."
    organization: str = Field(None, validate_default=True)
    "Name of the organization(/company) that the user belongs to."
    workspace: Optional[str] = None
    "Name of the workspace to use by default."
    project: Optional[str] = None
    "Name of the project to use by default."
    credentials: Optional[Credentials] = Field(
        default=None,
        validate_default=True,
    )
    "Authenticate via username/password instead of the device authorization code."
    skip_version_check: bool = False
    "Skip checking for updates."
    no_sse_connection: bool = False
    "Don't receive live updates from the SimAI API."
    https_proxy: Optional[AnyHttpUrl] = None
    "URL of the HTTPS proxy to use."
    tls_ca_bundle: Union[Literal["system", "unsecure-none"], PathLike, None] = None
    """
    Custom TLS CA certificate configuration. Possible values:

    * ``None``: use secure defaults
    * ``"system"``: uses system CA certificates (python >= 3.10)
    * A ``PathLike`` object: use a custom CA
    * ``"unsecure-none"``: no TLS certificate validation
    """

    @field_validator("url", mode="before")
    def clean_url(cls, url):
        if isinstance(url, bytes):
            url = url.decode()
        url = urlunparse(urlparse(url))
        if not url.endswith("/"):
            url = url + "/"
        return url

    def _auth_hash(self) -> str:
        hasher = hashlib.sha256()
        hasher.update(str(self.url).encode())
        if self.credentials:
            hasher.update(self.credentials.username.encode())
        config_file_profile = getattr(self, "_config_file_profile", None)
        if config_file_profile:
            hasher.update(config_file_profile.encode())
        return hasher.hexdigest()

    @field_validator("organization", mode="before")
    @classmethod
    def check_organization_exists(cls, val, info: ValidationInfo):
        """If organization is not set, either prompt the user or raise exception."""
        if not val:
            val = prompt_if_interactive(interactive=info.data["interactive"], name="organization")
        return val

    @field_validator("credentials", mode="before")
    @classmethod
    def creds_exist_when_non_interactive(cls, val, info: ValidationInfo):
        """If interactive mode is OFF and credentials are not set, raise exception."""
        if not info.data["interactive"] and not val:
            raise PydanticCustomError(
                "creds_missing_in_non_interactive",
                """Credentials should exist when interactive is false""",
            )
        return val



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import time
from typing import Callable, Optional

import requests
import sseclient

logger = logging.getLogger(__name__)


# take an optional last-event-id to be sent in request headers
EventSourceFactoryType = Callable[[Optional[str]], sseclient.SSEClient]


class ReconnectingSSERequestsClient:
    def __init__(
        self,
        event_source_factory: EventSourceFactoryType,
        char_enc: Optional[str] = None,
    ):
        self._sseclient = None
        self._event_source_factory = event_source_factory
        self._char_enc = char_enc
        self._last_event_id = None
        self._retry_timeout_sec = 3
        self._connect()

    def _connect(self):
        self._disconnect_client()
        logger.info(f"Will connect to SSE with last event id {self._last_event_id}.")
        event_source = self._event_source_factory(self._last_event_id)
        logger.info("Create SSEClient with event source.")
        self._sseclient = sseclient.SSEClient(event_source)

    def _disconnect_client(self):
        if self._sseclient is not None:
            try:
                self.close()
            except Exception:
                logger.error("Failed to close SSEClient ", exc_info=True)

    def close(self):
        self._sseclient.close()
        self._sseclient = None

    def events(self):
        while True:
            try:
                for event in self._sseclient.events():
                    yield event
                    self._last_event_id = event.id
            except (
                StopIteration,
                requests.exceptions.ChunkedEncodingError,
                requests.RequestException,
                EOFError,
            ) as e:
                logger.info(f"SSEClient disconnected: {e}")
                logger.info(f"Will try to reconnect after {self._retry_timeout_sec}s")
                time.sleep(self._retry_timeout_sec)
                self._connect()



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, TypeVar

from ansys.simai.core.data.base import (
    ERROR_STATES,
    PENDING_STATES,
    ComputableDataModel,
    Directory,
)
from ansys.simai.core.utils.numerical import cast_values_to_float

if TYPE_CHECKING:
    import ansys.simai.core.client

EXTRA_CALCULETTE_FIELDS = ["Area", "Normals", "Centroids"]

logger = logging.getLogger(__name__)


class GlobalCoefficientRequest(ABC, ComputableDataModel):
    """Creates the foundational request for subsequent Global Coefficients requests/inquiries."""

    def __init__(
        self,
        client: "ansys.simai.core.client.SimAIClient",
        directory: "Directory",
        fields: dict,
        project_id: str,
        gc_formula: str,
        sample_metadata: dict[str, Any],
        bc: list[str] = None,
        surface_variables: list[str] = None,
        **kwargs,
    ):
        super().__init__(client=client, directory=directory, fields=fields)

        self.project_id = project_id
        self._calculette_payload = self._compose_calculette_payload(
            gc_formula, sample_metadata, bc, surface_variables
        )

    @ComputableDataModel._failure_message.getter
    def _failure_message(self):
        return self.fields.get("error")

    def _compose_calculette_payload(
        self,
        gc_formula: str,
        sample_metadata: dict[str, Any],
        bc: list[str] = None,
        surface_variables: list[str] = None,
    ):
        """Composes the payload for a calculette request."""

        surface_vars_list = EXTRA_CALCULETTE_FIELDS
        if surface_variables:
            surface_vars_list += surface_variables

        return {
            "formula": gc_formula,
            "bc_list": bc if bc else [],
            "surface_field_list": [
                fd
                for fd in sample_metadata.get("surface").get("fields")
                if fd.get("name") in surface_vars_list
            ],
            "volume_field_list": [],
        }

    @abstractmethod
    def run(self):
        """Abstract method to perform the customized required process."""


GlobalCoefficientRequestType = TypeVar(
    "GlobalCoefficientRequestType", bound=GlobalCoefficientRequest
)


class GlobalCoefficientRequestDirectory(Directory[GlobalCoefficientRequestType]):
    """Provides methods for handling SSEs related to Global Coefficients."""

    def get(self, item_id: str) -> GlobalCoefficientRequestType:
        """Get a  by project ID.

        Args:
            item_id: ID of the Global Coefficient request.
        """

        return self._registry[item_id] if item_id not in self._registry else None

    def _handle_sse_event(self, data):
        gc_formula = data.get("target").get("formula")
        action = data.get("target").get("action")
        item_id: str = f"{data['target']['id']}-{action}-{gc_formula}"
        if item_id not in self._registry:
            logger.debug(
                f"{self.__class__.__name__}: Ignoring event for unknown object id {item_id}"
            )
            return
        item = self._registry[item_id]

        item._handle_job_sse_event(data)


class CheckGlobalCoefficient(GlobalCoefficientRequest):
    """Verifies a Global Coefficient formula."""

    def run(self):
        """Performs a check-formula request."""
        self._client._api.check_formula(self.project_id, self._calculette_payload)

    def _handle_job_sse_event(self, data):
        """Manage object's state according to SSE."""
        logger.debug(f"Handling SSE job event for {self._classname} id {self.id}")

        state: str = data.get("status")

        target = data.get("target")

        if state in PENDING_STATES:
            logger.debug(f"{self._classname} id {self.id} set status pending")
            self._set_is_pending()
        elif state == "successful":
            logger.debug(f"{self._classname} id {self.id} set status successful")
            self._set_is_over()
        elif state in ERROR_STATES:
            error_message = f"Verification of formula {target.get('formula')} failed with {data.get('reason', 'UNKNOWN ERROR')}"
            self.fields["error"] = error_message
            self._set_has_failed()
            logger.error(error_message)


class CheckGlobalCoefficientDirectory(GlobalCoefficientRequestDirectory[CheckGlobalCoefficient]):
    """Extends GlobalCoefficientRequestDirectory for verifying a Global Coefficient formula."""

    _data_model = CheckGlobalCoefficient


class ComputeGlobalCoefficient(GlobalCoefficientRequest):
    """Computes the result of a Global Coefficient formula."""

    def run(self):
        """Performs a compute-formula request."""
        self._client._api.compute_formula(self.project_id, self._calculette_payload)

    def _handle_job_sse_event(self, data):
        """Manage object's state according to SSE and store the result of the formula."""
        logger.debug(f"Handling SSE job event for {self._classname} id {self.id}")

        state: str = data.get("status")

        target = data.get("target")

        if state in PENDING_STATES:
            logger.debug(f"{self._classname} id {self.id} set status pending")
            self._set_is_pending()
        elif state == "successful":
            logger.debug(f"{self._classname} id {self.id} set status successful")
            self._result = cast_values_to_float(data.get("result", {}).get("value"))
            self._set_is_over()
        elif state in ERROR_STATES:
            error_message = f"Computation of global coefficient {target.get('formula')} failed with {data.get('reason', 'UNKNOWN ERROR')}"
            self.fields["error"] = error_message
            self._set_has_failed()
            logger.error(error_message)

    @property
    def result(self):
        """Get the result of the Global Coefficient formula."""
        return self._result if self.is_ready else None


class ComputeGlobalCoefficientDirectory(
    GlobalCoefficientRequestDirectory[ComputeGlobalCoefficient]
):
    """Extends GlobalCoefficientRequestDirectory for computing the result of a Global Coefficient formula."""

    _data_model = ComputeGlobalCoefficient



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import TYPE_CHECKING, Callable, Dict, Generic, List, Optional, TypeVar, Union

from ansys.simai.core.data.downloads import DownloadableResult
from ansys.simai.core.data.post_processings import PostProcessing

if TYPE_CHECKING:
    from ansys.simai.core.data.predictions import Prediction
    from ansys.simai.core.data.selections import Selection

from ansys.simai.core.errors import InvalidArguments, _foreach_despite_errors, _map_despite_errors

T = TypeVar("T", bound=PostProcessing)


class PPList(List, Generic[T]):
    """Provides a subclass of the :class:`list` class for storing postprocessings and adding a few shortcut methods.

    As a :class:`list` subclass, the ``PPList`` class supports any list operation.
    Its elements can be iterated on and accessed by index.
    """

    def __init__(self, selection: "Selection", post: Callable[["Prediction"], PostProcessing]):  # noqa: D107
        self._selection = selection
        # Even in case of errors, all post-processings will be queued.
        # but if some fail, an exception will be raised.
        post_processings = _map_despite_errors(lambda pred: post(pred), self._selection.predictions)
        super().__init__(post_processings)

    @property
    def data(self) -> Union[List[Dict[str, List]], List[DownloadableResult]]:
        """List containing the data of the underlying postprocessings.

        This is a blocking method, which returns once the data of all
        postprocessings is ready.
        """
        return [pp.data for pp in self]

    def wait(self):
        """Wait for all concerned postprocessings to finish."""
        _foreach_despite_errors(lambda pp: pp.wait(), self)


class ExportablePPList(PPList, Generic[T]):
    """Provides a subclass of the :class:`PPList` class for downloading the results of a group of postprocessings.

    As a :class:`list` subclass, the ``ExportablePPList`` class supports any list operation.
    Its elements can be iterated on and accessed by index.
    """

    def export(self, format: Optional[str] = "json") -> DownloadableResult:
        """Export the postprocessing results in the desired format.

        Accessing this property blocks until the data is ready.

        Args:
            format: format to exported data in. The default is ``'json'``.
                Options are ``'csv.zip'``, ``'json'``, and ``'xlsx'``.
                Note that the ``'csv.zip'`` option exports a ZIP file containing
                multiple CSV sheets.

        Returns:
            :class:`~ansys.simai.core.data.downloads.DownloadableResult` object for
            downloading the exported data into a file or access it in memory.
        """
        if format not in ["json", "csv.zip", "xlsx", "csv"]:
            raise InvalidArguments(
                f"Export format must be json, csv.zip, or xlsx (passed {format})."
            )
        if format == "csv":
            warnings.warn(
                "The ``csv`` option is being deprecated. Use the ``csv.zip`` option instead",
                PendingDeprecationWarning,
                stacklevel=1,
            )
        if len(self) < 1:
            raise InvalidArguments("Selection contains no exportable postprocessing.")
        # Wait for all concerned post-processings to finish (and raise if errors)
        self.wait()
        client = self[0]._client
        return DownloadableResult(
            client._api.post_processings_export_url(),
            client,
            request_method="POST",
            request_json_body={"ids": [pp.id for pp in self], "format": format},
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import math
from numbers import Number
from typing import Any, Optional, Tuple

from ansys.simai.core.errors import SimAIError

DEFAULT_COMPARISON_EPSILON = 10**-6

logger = logging.getLogger(__name__)


def is_number(value: Any):
    return isinstance(value, Number) and not math.isnan(value)


def is_smaller_with_tolerance(a: float, b: float, tolerance: Optional[float] = None):
    """Run the `less than`(<) comparison with a tolerance.

    The default for the tolerance is ``.000001``.

    If the difference between the two numbers is less
    than the tolerance, ``a`` is considered equal, not smaller.
    """
    if tolerance is None:
        tolerance = DEFAULT_COMPARISON_EPSILON
    return a < b - tolerance


def is_bigger_with_tolerance(a: float, b: float, tolerance: Optional[float] = None):
    """Run the `greater than` (>) comparison with a tolerance.

    The default for the tolerance is ``.000001``.

    If the difference between the two numbers is greater
    than the tolerance, ``b`` is considered equal, not larger.
    """
    if tolerance is None:
        tolerance = DEFAULT_COMPARISON_EPSILON
    return a > b + tolerance


def is_smaller_or_equal_with_tolerance(a: float, b: float, tolerance: Optional[float] = None):
    """Run the `less than or equal to` (<=) comparison with a tolerance.

    The default for the tolerance is ``.000001``.

    If the difference between the two numbers is smaller
    than the tolerance, ``b`` is considered equal.
    """
    return not is_bigger_with_tolerance(a, b, tolerance)


def is_bigger_or_equal_with_tolerance(a: float, b: float, tolerance: Optional[float] = None):
    """Run the `greater than or equal to` (>=) comparison with a tolerance.

    The default for the tolerance is ``.000001``.

    If the difference between the two numbers is smaller
    than the tolerance, ``b`` is considered equal.
    """
    return not is_smaller_with_tolerance(a, b, tolerance)


def is_equal_with_tolerance(a: float, b: float, tolerance: Optional[float] = None):
    """Compare the equality of two numbers with a tolerance.

    The default toleranance is ``.000001``.

    If the difference between the two numbers is smaller
    than the tolerance, they are considered equal.
    """
    if tolerance is None:
        tolerance = DEFAULT_COMPARISON_EPSILON

    return abs(a - b) <= tolerance


def convert_axis_and_coordinate_to_plane_eq_coeffs(
    axis: str, coordinate: float
) -> Tuple[float, float, float, float]:
    default_coeffs = [0, 0, 0, coordinate]
    if axis == "x":
        default_coeffs[0] = 1
    elif axis == "y":
        default_coeffs[1] = 1
    elif axis == "z":
        default_coeffs[2] = 1
    return tuple(default_coeffs)


def validate_tolerance_parameter(tolerance: Optional[Number]):
    if tolerance is None:
        return
    if not is_number(tolerance):
        raise TypeError(f"The tolerance argument must be a number >= 0 (passed {tolerance})")
    if tolerance < 0:
        raise ValueError(f"The tolerance argument must be a number >= 0 (passed {tolerance})")


# Recursively go through dict to cast values to float, so that we support NaN and inf
# Only use if all the values are expected to be numbers
def cast_values_to_float(data: Any):
    if isinstance(data, dict):
        return {k: cast_values_to_float(v) for k, v in data.items()}
    else:
        try:
            return float(data)
        except ValueError as err:
            raise SimAIError(f"received invalid float value: {data!r}") from err



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.model_configuration import ModelConfiguration
from ansys.simai.core.errors import InvalidArguments


class Model(ComputableDataModel):
    """Training model representation."""

    def __repr__(self) -> str:
        return f"<Model: {self.id}>"

    @property
    def project_id(self) -> str:
        """The ID of the project where the model exists."""
        return self.fields["project_id"]

    @property
    def configuration(self) -> ModelConfiguration:
        """Build configuration of a model."""
        return ModelConfiguration._from_payload(
            project=self._client.projects.get(self.fields["project_id"]),
            **self.fields["configuration"],
        )


class ModelDirectory(Directory[Model]):
    """Provides a collection of methods related to building models."""

    _data_model = Model

    def get(self, model_id: str) -> Model:
        """[Do not use] Get a model by project ID.

        Args:
            model_id: ID of the model.
        """

        raise NotImplementedError("The method 'get' of the class Model is not implemented yet.")

    def build(
        self,
        configuration: ModelConfiguration,
        dismiss_data_with_fields_discrepancies: bool = False,
        dismiss_data_with_volume_overflow: bool = False,
    ):
        """Launches a build given a configuration.

        Args:
            configuration: a ModelConfiguration object that contains the properties to be used in the build
            dismiss_data_with_fields_discrepancies: set to True for omitting data with missing properties
            dismiss_data_with_volume_overflow: set to True for omitting data outside the Domain of Analysis

        Example:
            Use a previous configuration for a new build in the same project

            .. code-block:: python

                a_project = simai.projects.get("project_A")

                build_conf = a_project.last_model_configuration

                new_model = simai.models.build(build_conf)

            Use a previous configuration for a new build in another project

            .. code-block:: python

                a_project = simai.projects.get("project_A")

                build_conf = a_project.last_model_configuration

                b_project = simai.projects.get("project_B")

                # set the id of b_project as the project_id of the configuration
                build_conf.project = b_project

                new_model = simai.models.build(build_conf)

        """
        if not configuration.project:
            raise InvalidArguments("The model configuration does not have a project set")

        is_trainable = configuration.project.is_trainable()
        if not is_trainable:
            raise InvalidArguments(f"Cannot train model because: {is_trainable.reason}")

        return self._model_from(
            self._client._api.launch_build(
                configuration.project.id,
                configuration._to_payload(),
                dismiss_data_with_fields_discrepancies,
                dismiss_data_with_volume_overflow,
            )
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Dict, List, Tuple

from ansys.simai.core.api.mixin import ApiClientMixin


class TrainingDataPartClientMixin(ApiClientMixin):
    def create_training_data_part(
        self, training_data_id: str, name: str, extension: str
    ) -> Tuple[Dict[str, Any], str]:
        """Create a part under the given training data without uploading the data.

        Args:
            training_data_id: ID of the parent training data.
            name: Name of the part to create.
            extension: Extension of the file or part.

        Returns:
            Tuple containing the ``TrainingDataPart`` object and the upload ID
            to use for further requests.
        """
        post_data = {"name": name, "file_extension": extension}
        response = self._post(f"training-data/{training_data_id}/parts/", json=post_data)
        return (response["training_data_part"], response["upload_id"])

    def get_training_data_part(self, id: str) -> Dict[str, Any]:
        return self._get(f"training-data-parts/{id}")

    def complete_training_data_part_upload(
        self, id: str, upload_id: str, parts: List[Dict[str, Any]]
    ):
        self._post(
            f"training-data-parts/{id}/complete",
            json={"upload_id": upload_id, "parts": parts},
        )



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Callable, Iterable, List, TypeVar

import requests


class SimAIError(Exception):
    """Provides the base exception for all errors of the SimAI client.

    To catch any expected error that the client might throw, use this exception.
    """


class ApiClientError(SimAIError, requests.exceptions.HTTPError):
    """HTTP error from the SimAI API."""

    def __init__(self, message: str, response=None):
        super(ApiClientError, self).__init__(message, response=response)

    @property
    def status_code(self):  # noqa: D102
        if self.response is not None:
            return self.response.status_code


class NotFoundError(ApiClientError):
    """Required resource was not found on the server."""


class ConnectionError(SimAIError, requests.exceptions.ConnectionError):
    """Could not communicate with the server."""


class ConfigurationError(SimAIError):
    """Client could not be configured properly."""


class ConfigurationNotFoundError(ConfigurationError):
    """Configuration file does not exist."""


class InvalidConfigurationError(ConfigurationError, ValueError):
    """Given configuration is not valid."""


class ProcessingError(SimAIError):
    """Data could not be processed."""


class InvalidArguments(SimAIError, ValueError):
    """Invalid arguments were provided."""


class InvalidClientStateError(SimAIError):
    """Client's state is invalid."""


class InvalidServerStateError(SimAIError):
    """Server's state is invalid."""


class InvalidOperationError(SimAIError):
    """This operation is not possible in the current state."""


class MultipleErrors(SimAIError):
    """Multiple errors occurred."""

    def __init__(self, exceptions: List[SimAIError]):
        self.exceptions = exceptions


T = TypeVar("T")


def _map_despite_errors(
    function: Callable[[T], Any],
    iterable: Iterable[T],
):
    """Like the map() method, this method applies the function for
    each item in the iterable and returns the result. On an exception,
    it continue with the next items. At the end, it raises either the
    exception or the ``MultipleError`` exception.
    """
    results: List[T] = []
    errors: List[SimAIError] = []
    for item in iterable:
        try:
            res = function(item)
            results.append(res)
        except SimAIError as e:
            errors.append(e)
    if errors:
        if len(errors) == 1:
            raise errors[0]
        raise MultipleErrors(errors)
    return results


def _foreach_despite_errors(
    procedure: Callable[[T], None],
    iterable: Iterable[T],
):
    """This method applies the procedure for each item in the
    iterable. On an exception, it continues with the next items.
    At the end, it raises either the exception or the ``MultipleError``
    exception.
    """
    errors = []
    for item in iterable:
        try:
            procedure(item)
        except SimAIError as e:
            errors.append(e)
    if errors:
        if len(errors) == 1:
            raise errors[0]
        raise MultipleErrors(errors)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import List, Optional

from pydantic import ValidationError

from ansys.simai.core import __version__
from ansys.simai.core.api.client import ApiClient
from ansys.simai.core.data.geometries import GeometryDirectory
from ansys.simai.core.data.global_coefficients_requests import (
    CheckGlobalCoefficientDirectory,
    ComputeGlobalCoefficientDirectory,
)
from ansys.simai.core.data.models import ModelDirectory
from ansys.simai.core.data.optimizations import (
    OptimizationDirectory,
    _OptimizationTrialRunDirectory,
)
from ansys.simai.core.data.post_processings import PostProcessingDirectory
from ansys.simai.core.data.predictions import PredictionDirectory
from ansys.simai.core.data.projects import Project, ProjectDirectory
from ansys.simai.core.data.training_data import TrainingDataDirectory
from ansys.simai.core.data.training_data_parts import TrainingDataPartDirectory
from ansys.simai.core.data.types import Path
from ansys.simai.core.data.workspaces import Workspace, WorkspaceDirectory
from ansys.simai.core.errors import (
    InvalidClientStateError,
    InvalidConfigurationError,
    MultipleErrors,
    NotFoundError,
    SimAIError,
)
from ansys.simai.core.utils.config_file import get_config
from ansys.simai.core.utils.configuration import ClientConfig
from ansys.simai.core.utils.misc import notify_if_package_outdated
from ansys.simai.core.utils.typing import steal_kwargs_type_on_method

logger = logging.getLogger(__name__)


class SimAIClient:
    """Provides the client for communicating with the SimAI API.

    For keyword arguments, see the :class:`~ansys.simai.core.utils.configuration.ClientConfig` class.

    Example:
        .. code-block:: python

            from ansys.simai.core import SimAIClient

            simai = SimAIClient(
                organization="company_name", https_proxy="http://company_proxy:3128"
            )
    """

    @steal_kwargs_type_on_method(ClientConfig)
    def __init__(self, **kwargs):
        try:
            config = ClientConfig(**kwargs)
        except ValidationError as pydandic_exc:
            raise InvalidConfigurationError(pydandic_exc) from None

        api_client_class = getattr(config, "_api_client_class_override", ApiClient)
        self._api = api_client_class(simai_client=self, config=config)
        self._check_gc_formula_directory = CheckGlobalCoefficientDirectory(client=self)
        self._compute_gc_formula_directory = ComputeGlobalCoefficientDirectory(client=self)
        self._geometry_directory = GeometryDirectory(client=self)
        self._optimization_directory = OptimizationDirectory(client=self)
        self._optimization_trial_run_directory = _OptimizationTrialRunDirectory(client=self)
        self._post_processing_directory = PostProcessingDirectory(client=self)
        self._project_directory = ProjectDirectory(client=self)
        self._model_directory = ModelDirectory(client=self)
        self._workspace_directory = WorkspaceDirectory(client=self)
        self._prediction_directory = PredictionDirectory(client=self)
        self._training_data_directory = TrainingDataDirectory(client=self)
        self._training_data_part_directory = TrainingDataPartDirectory(client=self)
        self._current_workspace = None
        self._current_project = None
        if config.workspace is not None:
            self.set_current_workspace(config.workspace)
        if config.project is not None:
            self.set_current_project(config.project)

        if not config.skip_version_check:
            self._check_for_new_version()

    @property
    def current_workspace(self) -> Workspace:
        """Workspace currently used by the SimAI client.

        Note:
            You should not set the workspace directly. Instead, use the :meth:`set_current_workspace`
            method, which uses the workspace name and ensures that the workspace exists.
        """
        if self._current_workspace is None:
            raise InvalidClientStateError("Current workspace is not set.")
        return self._current_workspace

    @current_workspace.setter
    def current_workspace(self, workspace: Workspace):
        self._current_workspace = workspace

    def set_current_workspace(self, name: str):
        """Set the current workspace for the SimAI client.

        Args:
            name: Name of the workspace that the client should switch to.

        Example:
            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config(workspace="old plane")
                simai.geometries.list()  # will list geometries belonging to the "old plane" workspace
                simai.set_current_workspace("new plane")
                simai.geometries.list()  # will list geometries belonging to the "new plane" workspace
        """
        try:
            # Ensure the workspace exists
            workspace = self.workspaces.get(name=name)
            logger.info(f"Workspace set to {name}.")
        except NotFoundError:
            raise InvalidConfigurationError(
                f"""Configured workspace {name} does not exist on the server.
{self._available_workspaces_string}"""
            ) from None
        self._current_workspace = workspace

    @property
    def _available_workspaces_string(self):
        available_workspaces = [workspace.name for workspace in self.workspaces.list()]
        return f"Available workspaces are: {available_workspaces}."

    @property
    def current_project(self) -> Project:
        """Project currently used by the SimAPI client.

        Note:
            You should not set the project directly. Instead, use the :meth:`set_current_project`
            method, which uses the project name and ensures that the project exists.
        """
        if self._current_project is None:
            raise InvalidClientStateError("Current project is not set.")
        return self._current_project

    @current_project.setter
    def current_project(self, project: Project):
        self._current_project = project

    def set_current_project(self, name: str):
        """Set the current project for the SimAI client.

        This method affects how some methods related to projects or associated
        data behave.

        Args:
            name: Name of the project that the client should switch to.
        """
        try:
            # Ensure the project exists
            project = self.projects.get(name=name)
            logger.info(f"Project set to {name}")
        except NotFoundError:
            raise InvalidConfigurationError(
                f"""Configured project {name} does not exist on the server.
{self._available_projects_string}"""
            ) from None
        self._current_project = project

    @property
    def _available_projects_string(self):
        available_projects = [project.name for project in self.projects.list()]
        return f"Available projects are: {available_projects}."

    @property
    def geometries(self):
        """Representation of all geometries on the server.
        For more information, see :ref:`geometries`.
        """
        return self._geometry_directory

    @property
    def optimizations(self):
        """Representation of all optimizations on the server.
        For more information, see :ref:`optimizations`.
        """
        return self._optimization_directory

    @property
    def training_data(self):
        """Representation of all training data on the server.
        For more information, see :ref:`training_data`.
        """
        return self._training_data_directory

    @property
    def training_data_parts(self):
        """Representation of all training data parts on the server.
        For more information, see :ref:`training_data_parts`.
        """
        return self._training_data_part_directory

    @property
    def predictions(self):
        """Representation of all predictions on the server.
        For more information, see :ref:`predictions`.
        """
        return self._prediction_directory

    @property
    def post_processings(self):
        """Representation of all postprocessings on the server.
        For more information, see :ref:`post_processings`.
        """
        return self._post_processing_directory

    @property
    def projects(self):
        """Representation of all projects on the server.
        For more information, see :ref:`projects`.
        """
        return self._project_directory

    @property
    def workspaces(self):
        """Representation of all workspaces on the server.
        For more information, see :ref:`workspaces`.
        """
        return self._workspace_directory

    @property
    def models(self):
        """Representation of all models on the server.
        For more information, see :ref:`models`.
        """
        return self._model_directory

    @classmethod
    def from_config(
        cls,
        profile: str = "default",
        path: Optional[Path] = None,
        **kwargs,
    ) -> "SimAIClient":
        """Initializes a SimAI client by reading a configuration file.

        You can provide the path of the configuration file to load. If no path is
        given, this method looks at the default locations.

        For more information, see :ref:`Configuration file<config_file>`.

        You can use ``kwargs`` to override part of the configuration.

        Args:
            profile: Profile to load from the configuration file. The default profile
                is loaded if no profile is provided.
            path: Path for the configuration file.
            **kwargs: Additional arguments to pass to the SimAI client.

        Returns:
            Configured client.

        Raises:
            ConfigurationNotFoundError: No configuration file was found at the given location
                or in the default profile if no path was given.
            InvalidConfigurationError: Configuration is invalid or incomplete.

        Example:
            Create the client after setting up your :ref:`configuration file.<config_file>`.

            .. code-block:: python

                import ansys.simai.core

                simai = ansys.simai.core.from_config()

        Note:
            The default paths are only supported on Unix systems.
        """
        return cls(**get_config(path, profile, **kwargs))

    def wait(self) -> None:
        """Wait for all ongoing operations on locally known predictions and postprocessings
        to finish.


        Raises:
            :exception:`SimAIError`: If something went wrong on an operation.
            :exception:`MultipleErrors`: If things went wrong on multiple operations.
        """
        errors: List[Exception] = []
        for directory in [self.predictions, self.post_processings]:
            logger.debug(f"waiting on {directory}")
            for local_object in directory._all_objects():
                try:
                    logger.debug(f"Waiting on {local_object._classname} id {local_object.id}")
                    local_object.wait()
                except Exception as e:
                    errors.append(e)
        if errors:
            if len(errors) == 1:
                raise errors[0]
            raise MultipleErrors(errors)

    def _check_for_new_version(self) -> None:
        try:
            latest_version = self._api._get("https://pypi.org/pypi/ansys-simai-core/json")["info"][
                "version"
            ]
        except (SimAIError, KeyError) as e:
            logger.debug(f"Could not query package version on pypi: {e}")
            return None
        notify_if_package_outdated("ansys-simai-core", __version__, latest_version)


from_config = SimAIClient.from_config



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Callable, Optional

from ansys.simai.core.utils.numerical import is_equal_with_tolerance


class _ToleranceGroup:
    """Provides a group of approximately equal values,
    as returned by ``itertools.groupby`` when using ``_ToleranceGrouper``.

    Note that with the way ``itertools.groupby`` works,
    accessing the center should be done only after accessing (iterating) the bucket's elements.
    """

    def __init__(
        self,
        initial_value: float,
        forced_central_value: Optional[float],
        tolerance: Optional[float],
    ):
        self.min = initial_value
        self.center = initial_value
        self.forced_central_value = forced_central_value
        self.tolerance = tolerance
        self.center_is_fixed = False

    def matches(self, new_value):
        if not is_equal_with_tolerance(new_value, self.center, self.tolerance):
            # The new value is just too far
            return False
        if new_value == self.forced_central_value:
            # Accept forced value to current bucket only if it's at reach for min
            # (else, refuse new value in this bucket)
            return is_equal_with_tolerance(new_value, self.min, self.tolerance)
        # Value is at reach of center (and not forced to be a new center):
        # it's all good
        return True

    def collect_value(self, new_value):
        # Check if the new value can be its center:
        # it must be at reach of bucket min,
        # and current center must not be fixed
        if not self.center_is_fixed and is_equal_with_tolerance(
            self.min, new_value, self.tolerance
        ):
            self.center = new_value
        # Fix the center of this bucket if it's the forced center
        if new_value == self.forced_central_value:
            self.center_is_fixed = True


class _ToleranceGrouper:
    """Provides a grouping class destined to be used by ``itertools.groupby`` to create
    groups of approximately equal value (according to tolerance).

    The ``_ToleranceGrouper`` class is meant to be passed as `key` argument for ``itertools.groupby``.
    It creates buckets defined by a central value and contains all values around it, plus or minus
    the tolerance. Each bucket is an instance of the ``_ToleranceGroup`` class.

    Args:
        key_func: Optional function computing a key value for each passed element.
            If no element is passed, the element itself is used.
        tolerance: Optional tolerance. The default is ``10**-6``.
        forced_central_value: Optional value that is forced to be at the
            center of its group. This means that the group spans maximally
            from forced_central_value - tolerance to forced_central_value + tolerance.
    """

    def __init__(
        self,
        key_func: Optional[Callable[[Any], float]] = None,
        tolerance: Optional[float] = None,
        forced_central_value: Optional[float] = None,
    ):
        if key_func is None:
            # if none passed, use identity function
            self.key_func = lambda x: x
        else:
            self.key_func = key_func
        self.tolerance = tolerance
        self.forced_central_value = forced_central_value
        self.current_bucket: Optional[_ToleranceGroup] = None

    def __call__(self, item: Any) -> float:
        new_value: float = self.key_func(item)
        if self.current_bucket is None or not self.current_bucket.matches(new_value):
            # Create a new bucket
            self.current_bucket = _ToleranceGroup(
                new_value,
                forced_central_value=self.forced_central_value,
                tolerance=self.tolerance,
            )
        self.current_bucket.collect_value(new_value)

        # Return bucket, which is used by groupby
        # to group all values together
        return self.current_bucket



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.simai.core.data.base import DataModel, Directory, UploadableResourceMixin


class TrainingDataPart(UploadableResourceMixin, DataModel):
    """Provides the local representation of a training data part object."""

    def __repr__(self) -> str:
        return f"<TrainingDataPart: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of the file."""
        return self.fields["name"]

    @property
    def size(self) -> int:
        """Size of the file in bytes."""
        return self.fields["size"]


class TrainingDataPartDirectory(Directory[TrainingDataPart]):
    """Provides the collection of methods related to training data parts.

    This class is accessed through ``client.training_data_parts``.
    """

    _data_model = TrainingDataPart

    def get(self, id: str) -> TrainingDataPart:
        """Get a :class:`TrainingDataPart` object from the server."""
        return self._model_from(self._client._api.get_training_data_part(id))



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
from typing import Any, BinaryIO, Dict, List, Optional, Union
from urllib.parse import quote

from ansys.simai.core.api.mixin import ApiClientMixin
from ansys.simai.core.data.types import File, MonitorCallback

logger = logging.getLogger(__name__)


class GeometryClientMixin(ApiClientMixin):
    """Provides the client for the Geometry ("/geometries/") part of the API."""

    def geometries(self, workspace_id: str, filters: Optional[Dict[str, Any]] = None):
        """Get list of all geometries."""
        if filters is not None:
            params = {"filters": json.dumps(filters), "workspace": workspace_id}
        else:
            params = {"workspace": workspace_id}
        return self._get("geometries/", params=params)

    def get_geometry(self, geometry_id: str):
        """Get information on a single geometry.

        Args:
            geometry_id: ID of the geometry.
        """
        return self._get(f"geometries/{geometry_id}")

    def get_geometry_by_name(self, name: str, workspace_id: str):
        """Get information on a single geometry by name instead of ID.

        Args:
            name: Name of the geometry.
            workspace_id: ID of the workspace that the geometry belongs to.
        """
        return self._get(f"geometries/name/{quote(name)}", params={"workspace": workspace_id})

    def delete_geometry(self, geometry_id: str):
        """Delete a single geometry.

        All objects associated with that geometry are also deleted.

        Args:
            geometry_id: ID of the geometry.
        """
        # TODO: Have user confirm or delete confirmation from API ?
        return self._delete(
            f"geometries/{geometry_id}",
            params={"confirm": True},
            return_json=False,
        )

    def update_geometry(
        self,
        geometry_id: str,
        name: Optional[str] = None,
        metadata: Optional[dict] = None,
    ):
        """Update the information for a given geometry.

        Args:
            geometry_id: ID of the geometry.
            name: New name to give to the geometries.
            metadata: Metadata to update the geometry with.
        """
        request_json = {}
        if name is not None:
            request_json["name"] = name
        if metadata is not None:
            request_json["metadata"] = metadata
        self._patch(f"geometries/{geometry_id}", json=request_json, return_json=False)

    def create_geometry(
        self,
        workspace_id: str,
        name: Optional[str] = None,
        extension: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """Create a geometry without pushing the data.

        Args:
            workspace_id: ID of the workspace to assign the geometry to.
            name: Name to give to the geometry.
            extension: Extension to give to the file.
            metadata: Metadata to apply to the geometry on creation.

        Returns:
           Tuple containing the geometry object and a 'presigned post'
           dictionary, which contains the URL to upload the data and fields to
           that was included in the request.
        """
        post_data = {
            "name": name,
            "workspace": workspace_id,
            "file_extension": extension,
        }
        if metadata:
            post_data.update({"metadata": json.dumps(metadata)})
        response = self._post(
            "geometries/",
            json=post_data,
        )
        return (response["geometry"], response["upload_id"])

    def complete_geometry_upload(self, id: str, upload_id: str, parts: List[Dict[str, Any]]):
        self._post(f"geometries/{id}/complete", json={"upload_id": upload_id, "parts": parts})

    def download_geometry(
        self,
        geometry_id: str,
        file: Optional[File] = None,
        monitor_callback: Optional[MonitorCallback] = None,
    ) -> Union[None, BinaryIO]:
        """Download the input geometry into the file at the given path.

        Args:
            geometry_id: ID of the geometry to download.
            file: Binary file-object or the path of the file to put the content into.
            monitor_callback: Function or method to pass the ``bytes_read`` delta to.
                This delta can be used to monitor progress.
        """
        return self.download_file(f"geometries/{geometry_id}/download", file, monitor_callback)

    def get_geometry_predictions(self, geometry_id: str):
        """Get predictions associated with a geometry.

        Args:
            geometry_id: ID of the geometry.
        """
        return self._get(f"geometries/{geometry_id}/predictions")



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import os
import random
import threading
import time
import webbrowser
from datetime import datetime, timedelta, timezone
from typing import Any, ClassVar, Optional
from urllib.parse import urljoin

import requests
from filelock import FileLock
from pydantic import BaseModel, ValidationError, computed_field, model_validator
from requests.auth import AuthBase
from requests.models import PreparedRequest

from ansys.simai.core.errors import ApiClientError
from ansys.simai.core.utils.configuration import ClientConfig, Credentials
from ansys.simai.core.utils.files import get_cache_dir
from ansys.simai.core.utils.requests import handle_response

logger = logging.getLogger(__name__)

# polling can't be faster or auth server returns HTTP 400 Slow down
DEVICE_AUTH_POLLING_INTERVAL = 5


class _AuthTokens(BaseModel):
    """Represents the OIDC tokens received from the auth server."""

    # Time buffer so we treat nearly invalid tokens as invalid.
    # Random so multiple processes don't refresh tokens at the same time.
    EXPIRATION_BUFFER: ClassVar = random.randrange(5, 15)  # noqa: S311

    access_token: str
    expiration: datetime
    refresh_expiration: datetime
    refresh_token: str
    # ...  (unused fields removed)

    @model_validator(mode="before")
    @classmethod
    def expires_in_to_datetime(cls, data: Any) -> dict:
        assert isinstance(data, dict)
        if "expiration" not in data:
            # We want to store "expiration" but API responses contains "expires_in"
            now = datetime.now(timezone.utc)
            data["expiration"] = now + timedelta(seconds=int(data["expires_in"]))
            data["refresh_expiration"] = now + timedelta(seconds=int(data["refresh_expires_in"]))
        return data

    def is_token_expired(self):
        return self.expires_in < self.EXPIRATION_BUFFER

    def is_refresh_token_expired(self):
        return self.refresh_expires_in < self.EXPIRATION_BUFFER

    @computed_field
    @property
    def expires_in(self) -> float:
        return (self.expiration - datetime.now(timezone.utc)).total_seconds()

    @computed_field
    @property
    def refresh_expires_in(self) -> float:
        return (self.refresh_expiration - datetime.now(timezone.utc)).total_seconds()


class _AuthTokensRetriever:
    """Retrieve tokens via ``get_tokens()``.
    It handles caching and the various auth token sources.
    """

    # Time buffer to refresh the refresh_token before it becomes invalid.
    # Random so multiple processes don't refresh the token at the same time
    REFRESH_BUFFER = random.randrange(50, 120)  # noqa: S311

    def __init__(
        self,
        credentials: Optional["Credentials"],
        session: requests.Session,
        auth_cache_hash: str,
        realm_url: str,
    ) -> None:
        self.credentials = credentials
        self.session = session
        self.token_url = f"{realm_url}/protocol/openid-connect/token"
        self.device_auth_url = f"{realm_url}/protocol/openid-connect/auth/device"
        self.refresh_timer = threading.Timer(0, lambda: None)
        self.cache_file_path = str(get_cache_dir() / f"tokens-{auth_cache_hash}.json")

    def _get_token_from_cache(self) -> Optional[_AuthTokens]:
        try:
            with open(self.cache_file_path, "r") as f:
                return _AuthTokens.model_validate_json(f.read())
        except (IOError, json.JSONDecodeError, ValidationError) as e:
            logger.info(f"Could not read auth token from cache: {e}")
        return None

    def _request_token_direct_grant(self) -> "_AuthTokens":
        logger.debug("request authentication tokens via direct grant")
        assert self.credentials
        request_params = {
            "client_id": "sdk",
            "grant_type": "password",
            "scope": "openid",
            **self.credentials.model_dump(),
        }
        return _AuthTokens(
            **handle_response(self.session.post(self.token_url, data=request_params))
        )

    def _request_token_device_auth(self) -> "_AuthTokens":
        logger.debug("request authentication tokens via device auth")
        auth_codes = handle_response(
            self.session.post(self.device_auth_url, data={"client_id": "sdk", "scope": "openid"})
        )
        print(  # noqa: T201
            f"Go to {auth_codes['verification_uri']} and enter the code {auth_codes['user_code']}"
        )
        webbrowser.open(auth_codes["verification_uri_complete"])
        # loop will exit when auth server returns "400 Device code is expired"
        while True:
            time.sleep(DEVICE_AUTH_POLLING_INTERVAL)
            validation = self.session.post(
                self.token_url,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "client_id": "sdk",
                    "device_code": auth_codes["device_code"],
                },
            )
            if b"authorization_pending" not in validation.content:
                return _AuthTokens(**handle_response(validation))

    def _refresh_auth_token(self, refresh_token: str) -> Optional[_AuthTokens]:
        logger.debug("Refreshing authentication tokens.")
        request_params = {
            "client_id": "sdk",
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }
        try:
            return _AuthTokens(
                **handle_response(self.session.post(self.token_url, data=request_params))
            )
        except (requests.exceptions.ConnectionError, ApiClientError) as e:
            logger.error(f"Could not refresh authentication tokens: {e}")
            return None

    def _schedule_auth_refresh(self, refresh_expires_in: float):
        """Schedule authentication refresh to avoids refresh token expiring if the client is idle for a long time."""
        if refresh_expires_in <= self.REFRESH_BUFFER:
            # Skip scheduling: refresh token too close to expiration (max SSO session nearly reached)
            return
        self.refresh_timer.cancel()
        self.refresh_timer = threading.Timer(
            refresh_expires_in - self.REFRESH_BUFFER, self.get_tokens
        )
        self.refresh_timer.daemon = True
        self.refresh_timer.start()

    def get_tokens(self, force_refresh=False) -> _AuthTokens:
        auth = self._get_token_from_cache()
        if auth and not auth.is_token_expired() and not force_refresh:
            # fast path: avoid locking the tokens, return early
            self._schedule_auth_refresh(auth.refresh_expires_in)
            return auth
        with FileLock(self.cache_file_path + ".lock", timeout=600):
            # slow path: tokens are locked, will get refreshed
            auth = self._get_token_from_cache()  # might have changed while we waited the lock
            if auth and auth.is_refresh_token_expired():
                auth = None
            if auth and (force_refresh or auth.is_token_expired()):
                auth = self._refresh_auth_token(auth.refresh_token)
            if auth is None:
                if self.credentials:
                    auth = self._request_token_direct_grant()
                else:
                    auth = self._request_token_device_auth()
            with open(self.cache_file_path + "~", "w") as f:
                f.write(auth.model_dump_json())
            # rename is atomic
            os.replace(self.cache_file_path + "~", self.cache_file_path)
        self._schedule_auth_refresh(auth.refresh_expires_in)
        return auth


class Authenticator(AuthBase):
    def __init__(self, config: ClientConfig, session: requests.Session) -> None:
        self._session = session
        self._enabled = not getattr(config, "_disable_authentication", False)
        if not self._enabled:
            logger.debug("Disabling authentication logic.")
            return
        self._url_prefix = config.url
        # HACK: start with a slash to override the /v2/ on the api url
        self._realm_url = urljoin(str(config.url), "/auth/realms/simai")
        self._organization_name = config.organization
        self._refresh_timer = None
        auth_hash = config._auth_hash()
        self.tokens_retriever = _AuthTokensRetriever(
            config.credentials, session, auth_hash, self._realm_url
        )
        self.tokens_retriever.get_tokens(
            force_refresh=True
        )  # start fetching/refreshing auth tokens

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        """Call to prepare the requests.

        Args:
            request: Request to authenticate.

        Returns:
            Request with the authentication.
        """
        assert request.url
        request_host = request.url.split("://", 1)[-1]  # ignore protocol part
        if (
            self._enabled
            and request_host.startswith(self._url_prefix.host)
            and self._realm_url not in request.url
        ):
            # So the token doesn't expire during requests that upload a large amount of data
            is_request_multipart_data = "multipart/form_data" in request.headers.get(
                "Content-Type", ""
            )
            auth = self.tokens_retriever.get_tokens(force_refresh=is_request_multipart_data)
            request.headers["Authorization"] = f"Bearer {auth.access_token}"
            request.headers["X-Org"] = self._organization_name
        return request



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from json.decoder import JSONDecodeError
from typing import Literal, overload

import requests

from ansys.simai.core.data.types import JSON, APIResponse
from ansys.simai.core.errors import ApiClientError, NotFoundError

logger = logging.getLogger(__name__)


def handle_http_errors(response: requests.Response) -> None:
    """Raise an error if the response status code is an error.

    Args:
        response: Response to check for errors.

    Raises:
        NotFoundError: If the response is a 404 error.
        ApiClientError: If the response is a 4xx or 5xx error other than the 404 error.
    """
    logger.debug("Checking for HTTP errors.")
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        try:
            json_response = response.json()
        except (ValueError, JSONDecodeError):
            # raise the errors from None
            # as we want to ignore the JSONDecodeError
            if response.status_code == 404:
                raise NotFoundError("Not Found", response) from e
            else:
                raise ApiClientError(
                    f"{response.status_code} {response.reason}", response
                ) from None
        if isinstance(json_response, dict):
            message = (
                json_response.get("errors")
                or json_response.get("message")
                or json_response.get("status")
                or json_response.get("error_description")
                or response.reason
            )

            if response.status_code == 404:
                raise NotFoundError(f"{message}", response) from e
            else:
                error_message = f"{response.status_code} {message}"
                if resolution := json_response.get("resolution", ""):
                    error_message += f"\n{resolution}"
                raise ApiClientError(
                    error_message,
                    response,
                ) from e
        else:
            raise ApiClientError(f"{response.status_code}: {response.reason}", response) from e


@overload
def handle_response(response: requests.Response, return_json: Literal[True]) -> JSON: ...


@overload
def handle_response(
    response: requests.Response, return_json: Literal[False]
) -> requests.Response: ...


@overload
def handle_response(response: requests.Response, return_json: bool) -> APIResponse: ...


def handle_response(response: requests.Response, return_json: bool = True) -> APIResponse:
    """Handle HTTP errors and return the relevant data from the response.

    Args:
        response: Response to handle
        return_json: Whether to return the JSON content or the whole response.

    Returns:
        JSON dict of the response if :py:args:`return_json` is ``True`` or the raw
            :py:class:`requests.Response` otherwise.
    """
    handle_http_errors(response)

    logger.debug("Returning response.")
    if return_json:
        try:
            return response.json()
        except (ValueError, JSONDecodeError):
            logger.debug("Failed to read JSON response.")
            raise ApiClientError(
                "Expected a JSON response but did not receive one.", response
            ) from None
    else:
        return response



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import getpass
import logging
from typing import Any, Dict, Optional

from semver.version import Version

logger = logging.getLogger(__name__)


def prompt_for_input(name: str, hide_input: Optional[bool] = False):
    return input(f"{name}:") if not hide_input else getpass.getpass(f"{name}:")


def build_boundary_conditions(boundary_conditions: Optional[Dict[str, Any]] = None, **kwargs):
    bc = boundary_conditions if boundary_conditions else {}
    bc.update(**kwargs)
    if not bc:
        raise ValueError("No boundary condition was specified.")
    return bc


def dict_get(obj: dict, *keys: str, default=None):
    """Get the requested key of the dictionary or opt to the default."""
    for k in keys:
        obj = obj.get(k, {}) or {}
    return obj or default


def notify_if_package_outdated(package: str, current_version: str, latest_version: str):
    try:
        version_current = Version.parse(current_version)
        version_latest = Version.parse(latest_version)
    except ValueError as e:
        logger.debug(f"Could not parse package version: {e}")
    else:
        if version_current < version_latest:
            warn_template = (
                f"A new version of {package} is %s. "
                "Upgrade to get new features and ensure compatibility with the server."
            )
            if (
                version_current.major < version_latest.major
                or version_current.minor < version_latest.minor
            ):
                logger.critical(warn_template % "required")
            else:
                logger.warning(warn_template % "available")
